from .resolver import (
    resolve_single_tensor,
    resolve_target_names,
)
from .validation import (
    ensure_mapping_payload,
    parse_torch_dtype,
    require_expr,
    require_nonempty_string,
    require_numeric,
    require_same_shape_dtype_device,
    require_same_shape_dtype_device3,
    validate_payload_keys,
)
from .types import (
    StateDictLike,
    StateDictProvider,
    TransformError,
)
from .scalar_comparison import (
    ScalarComparison,
    parse_scalar_comparison,
)
from .refs import (
    TensorRef,
    format_tensor_ref,
    must_model,
    parse_model_expr,
    parse_slice,
    select_tensor,
)
from .matching import (
    MatchError,
    StructuredMatch,
    StructuredPathMatcher,
)
from .name_mapping import (
    ResolvedMapping,
    match_expr_names,
    require_dest_missing,
    require_dest_present,
    resolve_name_mappings,
)
from .transform import (
    BaseTransform,
    BinaryMappingSpec,
    BinaryMappingTransform,
    CompiledTransform,
    DestinationPolicy,
    IteratingTransform,
    ResolvedTernaryMapping,
    TernaryMappingSpec,
    TernaryMappingTransform,
    TransformControl,
    TransformResult,
    TypedTransform,
    UnarySpec,
    UnaryTransform,
    REGISTRY,
    apply_transform,
    get_transform,
    list_transforms,
    register_transform,
)
from .declarative import (
    BinaryRefs,
    DeclarativeBinaryTransform,
    DeclarativeTernaryTransform,
    DeclarativeUnaryTransform,
    Docs,
    TernaryRefs,
    UnaryRefs,
)

from .expression import (
    Expression,
    ExpressionHelp,
    collect_expr_models,
    collect_ref_models,
    compile_assert_expr,
    compile_shape,
    compile_tensor_ref_expr,
    format_ref,
    get_assert_expr_help,
    get_assert_expr_names,
    register_assert_expr,
    require_mapping_assert_payload,
    resolve_matches,
    resolve_tensor_mappings,
    resolve_tensors,
)



from __future__ import annotations

import logging
from collections.abc import Iterator
from contextlib import contextmanager
from typing import Any

import typer
from rich.console import Console
from rich.panel import Panel

from .complete import (
    _collect_completion_candidates,
    _collect_payload_candidates,
    _completion_display_hook,
    _configure_readline_completion_bindings as _configure_readline_completion_bindings_impl,
    _infer_active_transform,
    _is_top_level_completion_position,
    _match_payload_candidates,
    _list_model_aliases,
    _payload_context,
    _render_completion_preview,
    _is_transform_payload_start,
)
from .history import _add_history_entry
from .parse import normalize_transform_specs, parse_transform_block

logger = logging.getLogger("brainsurgery")
console = Console()

try:
    import readline
except ImportError:  # pragma: no cover
    readline = None


def _configure_readline_completion_bindings() -> None:
    _configure_readline_completion_bindings_impl(readline)


@contextmanager
def _interactive_completion(
    *,
    top_level_candidates: list[str],
    lines: list[str],
    state_dict_provider: Any | None,
) -> Iterator[None]:
    if readline is None:
        yield
        return

    previous_completer = readline.get_completer()
    previous_delims = readline.get_completer_delims()
    matches: list[str] = []

    def completer(text: str, state: int) -> str | None:
        nonlocal matches
        if state == 0:
            try:
                line_buffer = readline.get_line_buffer()
                begidx = readline.get_begidx()
                get_endidx = getattr(readline, "get_endidx", None)
                if callable(get_endidx):
                    endidx = get_endidx()
                else:
                    endidx = begidx + len(text)
            except Exception:
                line_buffer = ""
                begidx = 0
                endidx = 0

            if _is_top_level_completion_position(line_buffer, begidx):
                matches = [
                    candidate for candidate in top_level_candidates if candidate.startswith(text)
                ]
                if not matches and ":" in line_buffer:
                    active_transform = _infer_active_transform(lines, line_buffer)
                    payload_candidates = _collect_payload_candidates(
                        active_transform=active_transform,
                        state_dict_provider=state_dict_provider,
                    )
                    matches = _match_payload_candidates(
                        text=text,
                        line_buffer=line_buffer,
                        begidx=begidx,
                        endidx=endidx,
                        payload_candidates=payload_candidates,
                        active_transform=active_transform,
                        model_aliases=sorted(_list_model_aliases(state_dict_provider)),
                    )
            else:
                active_transform = _infer_active_transform(lines, line_buffer)
                payload_candidates = _collect_payload_candidates(
                    active_transform=active_transform,
                    state_dict_provider=state_dict_provider,
                )
                matches = _match_payload_candidates(
                    text=text,
                    line_buffer=line_buffer,
                    begidx=begidx,
                    endidx=endidx,
                    payload_candidates=payload_candidates,
                    active_transform=active_transform,
                    model_aliases=sorted(_list_model_aliases(state_dict_provider)),
                )
        if state < len(matches):
            return matches[state]
        return None

    _configure_readline_completion_bindings()

    try:
        readline.set_completer_delims(" \t\n")
        readline.set_completer(completer)
    except Exception:
        logger.debug("Could not configure readline completion", exc_info=True)

    try:
        yield
    finally:
        try:
            readline.set_completer(previous_completer)
            readline.set_completer_delims(previous_delims)
        except Exception:
            logger.debug("Could not restore readline completion", exc_info=True)


def prompt_interactive_transform(state_dict_provider: Any | None = None) -> list[dict[str, Any]] | None:
    console.print()
    console.print(
        Panel.fit(
            "\n".join(
                [
                    "[bold]Interactive mode[/bold]",
                    "Enter one transform as YAML, or a YAML list of transforms.",
                    "Finish input with an empty line.",
                    "Tab completion: top-level transforms, then payload keys/aliases/tensors/YAML syntax.",
                    "Special transforms: [cyan]help[/cyan], [cyan]exit[/cyan]",
                    "Example: [dim]copy: { from: ln_f.weight, to: ln_f_copy.weight }[/dim]",
                ]
            ),
            border_style="cyan",
        )
    )

    while True:
        lines: list[str] = []
        prompt = typer.style("brainsurgery> ", fg=typer.colors.CYAN, bold=True)
        continuation = typer.style("... ", fg=typer.colors.BRIGHT_BLACK)
        top_level_candidates = _collect_completion_candidates(state_dict_provider)

        with _interactive_completion(
            top_level_candidates=top_level_candidates,
            lines=lines,
            state_dict_provider=state_dict_provider,
        ):
            while True:
                try:
                    line = input(prompt)
                except KeyboardInterrupt:
                    console.print()
                    console.print("KeyboardInterrupt")
                    break
                except EOFError:
                    return None

                if line.strip() == "":
                    if not lines:
                        continue

                    block = "\n".join(lines)
                    try:
                        parsed = parse_transform_block(block)
                        _add_history_entry(block)
                        return parsed
                    except ValueError as exc:
                        logger.error("Interactive transform rejected: %s", exc)
                        console.print(f"[red]Rejected:[/red] {exc}")
                        console.print("[yellow]Try again.[/yellow]")
                        break

                lines.append(line)
                prompt = continuation



from __future__ import annotations

from dataclasses import dataclass

from ..core import BinaryMappingSpec, DestinationPolicy, UnarySpec
from ..core import ResolvedMapping, StateDictProvider, TensorRef, TransformError, must_model, parse_slice, select_tensor
from ..core import register_transform
from ..core import require_numeric
from ..core import BinaryRefs, DeclarativeBinaryTransform, Docs, DeclarativeUnaryTransform, UnaryRefs


@dataclass(frozen=True)
class ClampSpec(BinaryMappingSpec):
    min_value: float | None
    max_value: float | None


def _build_clamp_spec(
    from_ref: TensorRef, to_ref: TensorRef, payload: dict
) -> ClampSpec:
    min_value, max_value = _parse_bounds(payload, "clamp", TransformError)
    return ClampSpec(
        from_ref=from_ref, to_ref=to_ref, min_value=min_value, max_value=max_value
    )


def _clamp_apply(
    spec: ClampSpec, item: ResolvedMapping, provider: StateDictProvider
) -> None:
    src_sd = provider.get_state_dict(item.src_model)
    dst_sd = provider.get_state_dict(item.dst_model)
    src_view = select_tensor(src_sd[item.src_name], item.src_slice)
    dst_sd[item.dst_name] = src_view.clamp(
        min=spec.min_value, max=spec.max_value
    ).clone()


class ClampTransform(DeclarativeBinaryTransform[ClampSpec]):
    name = "clamp"
    error_type = TransformError
    spec_type = ClampSpec
    destination_policy = DestinationPolicy.MUST_NOT_EXIST
    allowed_keys = {"from", "to", "min", "max"}
    required_keys = {"from", "to"}
    docs = Docs(
        "Clamps source tensors into new destination tensors.",
        notes=("At least one of 'min' or 'max' is required.",),
        examples=("clamp: { from: x, to: x_clamped, min: -1.0, max: 1.0 }",),
    )
    refs = BinaryRefs(from_slice=True)
    spec_builder = staticmethod(_build_clamp_spec)
    apply_fn = staticmethod(_clamp_apply)


def _parse_bounds(
    payload: dict,
    op_name: str,
    error_type: type[TransformError],
) -> tuple[float | None, float | None]:
    min_value = (
        require_numeric(payload, op_name=op_name, key="min")
        if "min" in payload
        else None
    )
    max_value = (
        require_numeric(payload, op_name=op_name, key="max")
        if "max" in payload
        else None
    )
    if min_value is None and max_value is None:
        raise error_type(f"{op_name} requires at least one of: min, max")
    if min_value is not None and max_value is not None and min_value > max_value:
        raise error_type(f"{op_name}.min must be <= {op_name}.max")
    return min_value, max_value


register_transform(ClampTransform())


@dataclass(frozen=True)
class ClampInPlaceSpec(UnarySpec):
    min_value: float | None
    max_value: float | None


def _build_clamp_in_place_spec(
    target_ref: TensorRef, payload: dict
) -> ClampInPlaceSpec:
    min_value, max_value = _parse_bounds(payload, "clamp_", TransformError)
    return ClampInPlaceSpec(
        target_ref=target_ref, min_value=min_value, max_value=max_value
    )


def _clamp_in_place_apply(
    spec: ClampInPlaceSpec, name: str, provider: StateDictProvider
) -> None:
    model = must_model(spec.target_ref)
    sd = provider.get_state_dict(model)
    slice_spec = (
        parse_slice(spec.target_ref.slice_spec)
        if spec.target_ref.slice_spec is not None
        else None
    )
    view = select_tensor(sd[name], slice_spec)
    view.clamp_(min=spec.min_value, max=spec.max_value)
    sd.mark_write(name)


class ClampInPlaceTransform(DeclarativeUnaryTransform[ClampInPlaceSpec]):
    name = "clamp_"
    error_type = TransformError
    spec_type = ClampInPlaceSpec
    allowed_keys = {"target", "min", "max"}
    required_keys = {"target"}
    docs = Docs(
        "Clamps target tensors in-place.",
        notes=("At least one of 'min' or 'max' is required.",),
        examples=("clamp_: { target: x, min: -1.0, max: 1.0 }",),
    )
    refs = UnaryRefs(target_slice=True)
    spec_builder = staticmethod(_build_clamp_in_place_spec)
    apply_fn = staticmethod(_clamp_in_place_apply)


register_transform(ClampInPlaceTransform())



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Tuple

import torch

from .types import TransformError


Expr = str | list[str]


@dataclass(frozen=True)
class TensorRef:
    model: Optional[str]
    expr: Expr
    slice_spec: Optional[str] = None


def parse_model_expr(raw: object, default_model: Optional[str] = None) -> TensorRef:
    if isinstance(raw, list):
        if default_model is None:
            raise TransformError("missing model alias for structured reference")
        if not raw:
            raise TransformError("structured reference must be a non-empty list")
        if not all(isinstance(item, str) and item for item in raw):
            raise TransformError("structured reference must be a non-empty list of non-empty strings")
        return TensorRef(model=default_model, expr=raw, slice_spec=None)

    if not isinstance(raw, str) or not raw:
        raise TransformError("reference must be a non-empty string or a non-empty list of strings")

    parts = raw.split("::")
    if len(parts) == 1:
        if default_model is None:
            raise TransformError(f"missing model alias in reference: {raw!r}")
        return TensorRef(model=default_model, expr=parts[0], slice_spec=None)

    if len(parts) == 2:
        head, tail = parts
        if default_model is not None and _looks_like_slice(tail):
            return TensorRef(model=default_model, expr=head, slice_spec=tail)
        return TensorRef(model=head or default_model, expr=tail, slice_spec=None)

    if len(parts) == 3:
        head, expr, slice_spec = parts
        if not _looks_like_slice(slice_spec):
            raise TransformError(f"invalid slice syntax in reference: {raw!r}")
        model = head or default_model
        if model is None:
            raise TransformError(f"missing model alias in reference: {raw!r}")
        return TensorRef(model=model, expr=expr, slice_spec=slice_spec)

    raise TransformError(f"invalid reference syntax: {raw!r}")


def parse_slice(raw: str) -> Tuple[object, ...]:
    if not _looks_like_slice(raw):
        raise TransformError(f"invalid slice syntax: {raw!r}")

    inner = raw[1:-1].strip()
    if not inner:
        return tuple()

    parts = [part.strip() for part in inner.split(",")]
    if any(part == "" for part in parts):
        raise TransformError(f"invalid empty slice component in {raw!r}")

    return tuple(_parse_slice_component(part) for part in parts)


def _parse_slice_component(raw: str) -> object:
    if raw == ":":
        return slice(None, None, None)

    if ":" not in raw:
        return _parse_int(raw)

    parts = raw.split(":")
    if len(parts) not in (2, 3):
        raise TransformError(f"invalid slice component: {raw!r}")

    start = _parse_optional_int(parts[0])
    stop = _parse_optional_int(parts[1])
    step = _parse_optional_int(parts[2]) if len(parts) == 3 else None
    return slice(start, stop, step)


def select_tensor(tensor: torch.Tensor, slice_spec: Optional[Tuple[object, ...]]) -> torch.Tensor:
    if slice_spec is None:
        return tensor
    try:
        return tensor[slice_spec]
    except Exception as exc:  # pragma: no cover
        raise TransformError(
            f"failed to apply slice {slice_spec!r} to tensor with shape {tuple(tensor.shape)}"
        ) from exc


def _parse_int(raw: str) -> int:
    try:
        return int(raw)
    except ValueError as exc:
        raise TransformError(f"invalid integer in slice component: {raw!r}") from exc


def _parse_optional_int(raw: str) -> Optional[int]:
    return None if raw == "" else _parse_int(raw)


def _looks_like_slice(raw: str) -> bool:
    return raw.startswith("[") and raw.endswith("]")


def must_model(ref: TensorRef) -> str:
    if ref.model is None:
        raise TransformError(f"reference is missing model alias: {ref}")
    return ref.model


def _format_ref_expr(expr: Expr) -> str:
    if isinstance(expr, str):
        return expr
    return "[" + ", ".join(repr(part) for part in expr) + "]"


def format_tensor_ref(ref: TensorRef) -> str:
    model = must_model(ref)
    expr = _format_ref_expr(ref.expr)
    if ref.slice_spec is None:
        return f"{model}::{expr}"
    return f"{model}::{expr}::{ref.slice_spec}"


def _validate_expr_kind(
    *,
    expr: Expr,
    op_name: str,
    role: str,
) -> None:
    if isinstance(expr, str):
        if not expr:
            raise TransformError(f"{op_name} {role} regex must be non-empty")
        return

    if isinstance(expr, list):
        if not expr:
            raise TransformError(f"{op_name} structured {role} pattern must be non-empty")
        if not all(isinstance(item, str) and item for item in expr):
            raise TransformError(
                f"{op_name} structured {role} pattern must be a list of non-empty strings"
            )
        return

    raise TransformError(f"{op_name} {role} expression has invalid type: {type(expr).__name__}")


__all__ = [
    "TensorRef",
    "format_tensor_ref",
    "must_model",
    "parse_model_expr",
    "parse_slice",
    "select_tensor",
]



from .execution import execute_transform_pairs
from .arena import ProviderError
from .state_dicts import (
    ArenaStateDict,
    InMemoryStateDict,
)
from .providers import (
    BaseStateDictProvider,
    InMemoryStateDictProvider,
    create_state_dict_provider,
)
from .provider_utils import (
    find_alias_mapping,
    get_or_create_alias_state_dict,
    iter_alias_mappings,
    list_model_aliases,
    new_empty_state_dict,
    resolve_single_model_alias,
)
from .render import render_tree, summarize_tensor
from .plan import compile_plan
from .frontend import emit_line, set_output_emitter, use_output_emitter



from __future__ import annotations

from collections.abc import Callable
from contextlib import contextmanager

OutputEmitter = Callable[[str], None]

_output_emitter: OutputEmitter | None = None


def emit_line(message: str = "") -> None:
    emitter = _output_emitter
    if emitter is None:
        print(message)
        return
    emitter(message)


def set_output_emitter(emitter: OutputEmitter | None) -> OutputEmitter | None:
    global _output_emitter
    previous = _output_emitter
    _output_emitter = emitter
    return previous


@contextmanager
def use_output_emitter(emitter: OutputEmitter | None):
    previous = set_output_emitter(emitter)
    try:
        yield
    finally:
        set_output_emitter(previous)



from __future__ import annotations

import re
from pathlib import Path
from typing import Literal

from .plan import OutputSpec


def resolve_output_destination(
    output: OutputSpec,
    *,
    default_shard_size: str,
) -> tuple[Path, Literal["safetensors", "torch"], int | None]:
    path = output.path
    format_value = output.format
    shard_size = resolve_shard_size(output, default_shard_size=default_shard_size)

    if format_value is not None:
        if format_value == "safetensors":
            resolved_path, resolved_format = _resolve_explicit_safetensors_destination(path)
        elif format_value == "torch":
            resolved_path, resolved_format = _resolve_explicit_torch_destination(path)
        else:  # pragma: no cover
            raise RuntimeError(f"unsupported explicit output format: {format_value}")
    else:
        suffix = path.suffix.lower()
        if suffix == ".safetensors":
            resolved_path, resolved_format = path, "safetensors"
        elif suffix in {".pt", ".pth", ".bin"}:
            resolved_path, resolved_format = path, "torch"
        elif path.exists() and path.is_dir():
            resolved_path, resolved_format = path / "model.safetensors", "safetensors"
        elif suffix == "":
            resolved_path, resolved_format = path / "model.safetensors", "safetensors"
        else:
            raise RuntimeError(
                f"unsupported output format for {path}; use a directory, .safetensors, .pt, .pth, or .bin, "
                f"or specify output.format explicitly"
            )

    if shard_size is not None and resolved_format != "safetensors":
        raise RuntimeError("output.shard is only supported for safetensors output")

    return resolved_path, resolved_format, shard_size


def _resolve_explicit_safetensors_destination(path: Path) -> tuple[Path, Literal["safetensors"]]:
    if path.exists() and path.is_dir():
        return path / "model.safetensors", "safetensors"
    if path.suffix == "":
        return path / "model.safetensors", "safetensors"
    if path.suffix.lower() != ".safetensors":
        raise RuntimeError(
            f"output.format='safetensors' is incompatible with file path {path}; "
            f"use a directory or a .safetensors file"
        )
    return path, "safetensors"


def _resolve_explicit_torch_destination(path: Path) -> tuple[Path, Literal["torch"]]:
    if path.exists() and path.is_dir():
        raise RuntimeError("output.format='torch' requires a file path, not a directory")
    if path.suffix.lower() not in {".pt", ".pth", ".bin"}:
        raise RuntimeError(
            f"output.format='torch' requires a .pt, .pth, or .bin file path; got {path}"
        )
    return path, "torch"


def resolve_shard_size(output: OutputSpec, default_shard_size: str) -> int | None:
    raw = output.shard

    if raw is None:
        if _is_directory_style_output(output):
            raw = default_shard_size
        else:
            return None

    return parse_shard_size(raw)


def _is_directory_style_output(output: OutputSpec) -> bool:
    path = output.path

    if output.format == "torch":
        return False

    if output.format == "safetensors":
        if path.exists() and path.is_dir():
            return True
        return path.suffix == ""

    if path.exists() and path.is_dir():
        return True

    return path.suffix == ""


def parse_shard_size(raw: str | None) -> int | None:
    if raw is None or raw == "none":
        return None

    if not isinstance(raw, str) or not raw:
        raise RuntimeError("output.shard must be a non-empty string or 'none'")

    match = re.fullmatch(r"(?i)\s*(\d+)\s*(b|kb|mb|gb|tb)\s*", raw)
    if not match:
        raise RuntimeError(
            f"invalid output.shard value {raw!r}; expected values like 'none', '500MB', '5GB'"
        )

    value = int(match.group(1))
    unit = match.group(2).lower()

    multipliers = {
        "b": 1,
        "kb": 1024,
        "mb": 1024**2,
        "gb": 1024**3,
        "tb": 1024**4,
    }

    return value * multipliers[unit]


def resolve_sharded_output_directory(original_path: Path, resolved_path: Path) -> Path:
    if original_path.exists() and original_path.is_dir():
        return original_path
    if original_path.suffix == "":
        return original_path
    raise RuntimeError(
        "sharded safetensors output requires a directory-style output path, not a single file"
    )



from __future__ import annotations

from abc import abstractmethod
from collections.abc import MutableMapping
from typing import Any, Protocol

import torch


class TransformError(RuntimeError):
    pass


class StateDictLike(MutableMapping[str, torch.Tensor]):
    @abstractmethod
    def slot(self, key: str) -> Any:
        raise NotImplementedError

    @abstractmethod
    def bind_slot(self, key: str, slot: Any) -> None:
        raise NotImplementedError

    @abstractmethod
    def access_counts(self, key: str) -> dict[str, int]:
        raise NotImplementedError

    @abstractmethod
    def mark_write(self, key: str, count: int = 1) -> None:
        raise NotImplementedError


class StateDictProvider(Protocol):
    def get_state_dict(self, model: str) -> StateDictLike:
        ...

__all__ = ["StateDictLike", "StateDictProvider", "TransformError"]



from __future__ import annotations

import torch

from .refs import Expr
from .types import TransformError


_DTYPE_ALIASES: dict[str, torch.dtype] = {
    "float16": torch.float16,
    "half": torch.float16,
    "fp16": torch.float16,
    "bfloat16": torch.bfloat16,
    "bf16": torch.bfloat16,
    "float32": torch.float32,
    "float": torch.float32,
    "fp32": torch.float32,
    "float64": torch.float64,
    "double": torch.float64,
    "fp64": torch.float64,
    "uint8": torch.uint8,
    "int8": torch.int8,
    "int16": torch.int16,
    "short": torch.int16,
    "int32": torch.int32,
    "int": torch.int32,
    "int64": torch.int64,
    "long": torch.int64,
    "bool": torch.bool,
}


def parse_torch_dtype(
    raw: str,
    *,
    error_type: type[TransformError],
    op_name: str,
    field_name: str,
) -> torch.dtype:
    value = raw.strip().lower()
    try:
        return _DTYPE_ALIASES[value]
    except KeyError as exc:
        allowed = ", ".join(sorted(_DTYPE_ALIASES))
        raise error_type(
            f"{op_name}.{field_name} unsupported dtype {raw!r}; expected one of: {allowed}"
        ) from exc


def ensure_mapping_payload(payload: object, op_name: str) -> dict:
    if not isinstance(payload, dict):
        raise TransformError(f"{op_name} payload must be a mapping")
    return payload


def validate_payload_keys(
    payload: dict,
    *,
    op_name: str,
    allowed_keys: set[str],
    required_keys: set[str] | None = None,
) -> None:
    unknown = set(payload) - allowed_keys
    if unknown:
        raise TransformError(f"{op_name} received unknown keys: {sorted(unknown)}")

    if required_keys is None:
        required_keys = set()

    missing = required_keys - set(payload)
    if missing:
        missing_list = sorted(missing)
        if len(missing_list) == 1:
            raise TransformError(f"{op_name}.{missing_list[0]} is required")
        raise TransformError(f"{op_name} is missing required keys: {missing_list}")


def require_nonempty_string(payload: dict, *, op_name: str, key: str) -> str:
    value = payload.get(key)
    if not isinstance(value, str) or not value:
        raise TransformError(f"{op_name}.{key} must be a non-empty string")
    return value


def require_expr(payload: dict, *, op_name: str, key: str) -> Expr:
    value = payload.get(key)

    if isinstance(value, str):
        if not value:
            raise TransformError(
                f"{op_name}.{key} must be a non-empty string or a non-empty list of non-empty strings"
            )
        return value

    if isinstance(value, list):
        if not value:
            raise TransformError(
                f"{op_name}.{key} must be a non-empty string or a non-empty list of non-empty strings"
            )
        if not all(isinstance(item, str) and item for item in value):
            raise TransformError(
                f"{op_name}.{key} must be a non-empty string or a non-empty list of non-empty strings"
            )
        return value

    raise TransformError(
        f"{op_name}.{key} must be a non-empty string or a non-empty list of non-empty strings"
    )


def require_numeric(payload: dict, *, op_name: str, key: str) -> float:
    value = payload.get(key)
    try:
        return float(value)
    except (TypeError, ValueError) as exc:
        raise TransformError(f"{op_name}.{key} must be numeric") from exc


def require_same_shape_dtype_device(
    left: torch.Tensor,
    right: torch.Tensor,
    *,
    op_name: str,
    left_name: str,
    right_name: str,
) -> None:
    if left.shape != right.shape:
        raise TransformError(
            f"shape mismatch {op_name} {left_name} -> {right_name}: "
            f"{tuple(left.shape)} != {tuple(right.shape)}"
        )
    if left.dtype != right.dtype:
        raise TransformError(
            f"dtype mismatch {op_name} {left_name} -> {right_name}: "
            f"{left.dtype} != {right.dtype}"
        )
    if left.device != right.device:
        raise TransformError(
            f"device mismatch {op_name} {left_name} -> {right_name}: "
            f"{left.device} != {right.device}"
        )


def require_same_shape_dtype_device3(
    first: torch.Tensor,
    second: torch.Tensor,
    dest: torch.Tensor,
    *,
    op_name: str,
    first_name: str,
    second_name: str,
    dest_name: str,
    symbol: str,
) -> None:
    if first.shape != second.shape or first.shape != dest.shape:
        raise TransformError(
            f"shape mismatch {op_name} {first_name} {symbol} {second_name} -> {dest_name}: "
            f"{tuple(first.shape)} {symbol} {tuple(second.shape)} -> {tuple(dest.shape)}"
        )
    if first.dtype != second.dtype or first.dtype != dest.dtype:
        raise TransformError(
            f"dtype mismatch {op_name} {first_name} {symbol} {second_name} -> {dest_name}: "
            f"{first.dtype} {symbol} {second.dtype} -> {dest.dtype}"
        )
    if first.device != second.device or first.device != dest.device:
        raise TransformError(
            f"device mismatch {op_name} {first_name} {symbol} {second_name} -> {dest_name}: "
            f"{first.device} {symbol} {second.device} -> {dest.device}"
        )


__all__ = [
    "ensure_mapping_payload",
    "parse_torch_dtype",
    "require_expr",
    "require_nonempty_string",
    "require_numeric",
    "require_same_shape_dtype_device",
    "require_same_shape_dtype_device3",
    "validate_payload_keys",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import torch

from ..core import Expression, TransformError, collect_ref_models, compile_tensor_ref_expr, format_ref, register_assert_expr, require_mapping_assert_payload, resolve_tensors
from ..core import TensorRef
from ..core import StateDictProvider


@dataclass(frozen=True)
class IsZeroExpr:
    ref: TensorRef
    eps: float | None = None

    def evaluate(self, provider: StateDictProvider) -> None:
        for ref, tensor in resolve_tensors(self.ref, provider, op_name="iszero"):
            if self.eps is None:
                is_zero = bool(torch.all(tensor == 0).item())
            else:
                if tensor.is_complex():
                    diff = torch.abs(tensor.to(torch.complex128))
                else:
                    diff = torch.abs(tensor.to(torch.float64))
                is_zero = bool(torch.all(diff <= self.eps).item())

            if not is_zero:
                if self.eps is None:
                    raise TransformError(f"iszero failed: {format_ref(ref)} is not all zeros")
                raise TransformError(
                    f"iszero failed: {format_ref(ref)} is not all zeros within eps={self.eps}"
                )

    def collect_models(self) -> set[str]:
        return collect_ref_models(self.ref)


@register_assert_expr(
    "iszero",
    payload_kind="tensor-ref|mapping",
    allowed_keys={"of", "eps"},
    required_keys={"of"},
    description=(
        "Succeeds if selected tensor(s) are all zeros, "
        "or within optional absolute tolerance 'eps'."
    ),
)
def compile_iszero_expr(payload: Any, default_model: str | None) -> Expression:
    if isinstance(payload, dict):
        payload = require_mapping_assert_payload(
            payload,
            op_name="iszero",
            allowed_keys={"of", "eps"},
            required_keys={"of"},
        )
        raw_ref = payload["of"]
        raw_eps = payload.get("eps")
    else:
        raw_ref = payload
        raw_eps = None

    ref = compile_tensor_ref_expr(raw_ref, default_model, "iszero")

    if raw_eps is None:
        eps = None
    else:
        if isinstance(raw_eps, bool) or not isinstance(raw_eps, (int, float)):
            raise TransformError("iszero.eps must be a non-negative number")
        eps = float(raw_eps)
        if eps < 0:
            raise TransformError("iszero.eps must be a non-negative number")

    return IsZeroExpr(ref=ref, eps=eps)



from __future__ import annotations

from dataclasses import dataclass

from ..core import IteratingTransform, UnarySpec, UnaryTransform
from ..algorithms.phlora import PhloraSvdCache, compute_phlora_factors, reconstruct_phlora_rank, require_positive_rank
from ..core import ResolvedMapping, resolve_name_mappings
from ..core import TensorRef, must_model, parse_model_expr
from ..core import register_transform
from ..core import ensure_mapping_payload, require_expr, require_numeric, validate_payload_keys
from ..core import StateDictProvider, TransformError


class PhloraTransformError(TransformError):
    pass


@dataclass(frozen=True)
class PhloraSpec:
    source_ref: TensorRef
    factor_b_ref: TensorRef
    factor_a_ref: TensorRef
    rank: int
    delete_source: bool
    require_missing_outputs: bool

    def collect_models(self) -> set[str]:
        return {
            must_model(self.source_ref),
            must_model(self.factor_a_ref),
            must_model(self.factor_b_ref),
        }


@dataclass(frozen=True)
class ResolvedPhloraMapping:
    source_model: str
    source_name: str
    factor_a_model: str
    factor_a_name: str
    factor_b_model: str
    factor_b_name: str


class PhloraTransform(IteratingTransform[PhloraSpec, ResolvedPhloraMapping]):
    name = "phlora"
    error_type = PhloraTransformError
    spec_type = PhloraSpec
    allowed_keys = {
        "target",
        "target_a",
        "target_b",
        "rank",
        "delete_original",
        "require_missing_dest",
    }
    required_keys = {"target", "target_a", "target_b", "rank"}
    progress_desc = "Applying phlora transforms"
    help_text = (
        "Splits each target tensor into PHLoRA low-rank factors B and A.\n"
        "\n"
        "Inputs:\n"
        "  - target: source tensor expression\n"
        "  - target_a: destination expression for A = sqrt(S) * Vh\n"
        "  - target_b: destination expression for B = U * sqrt(S)\n"
        "  - rank: low-rank dimension r\n"
        "\n"
        "target/target_a/target_b support regex or structured expressions like other transforms.\n"
        "Mappings are resolved from target -> target_a and target -> target_b; both must cover "
        "the same source set.\n"
        "\n"
        "Options:\n"
        "  - delete_original (default: true)\n"
        "  - require_missing_dest (default: true)\n"
        "\n"
        "Example:\n"
        "  phlora: { target: '(.*)\\\\.weight', rank: 16, target_a: '\\\\1.a', target_b: '\\\\1.b' }"
    )

    def __init__(self) -> None:
        super().__init__()
        self._svd_cache = PhloraSvdCache()

    def completion_reference_keys(self) -> list[str]:
        return ["target", "target_a", "target_b"]

    def compile(self, payload: dict, default_model: str | None) -> PhloraSpec:
        payload = ensure_mapping_payload(payload, self.name)
        validate_payload_keys(
            payload,
            op_name=self.name,
            allowed_keys=self.allowed_keys,
            required_keys=self.required_keys,
        )

        source_ref = parse_model_expr(
            require_expr(payload, op_name=self.name, key="target"),
            default_model=default_model,
        )
        source_model_default = source_ref.model if source_ref.model is not None else default_model
        factor_a_ref = parse_model_expr(
            require_expr(payload, op_name=self.name, key="target_a"),
            default_model=source_model_default,
        )
        factor_b_ref = parse_model_expr(
            require_expr(payload, op_name=self.name, key="target_b"),
            default_model=source_model_default,
        )

        self.validate_refs(source_ref, factor_b_ref, factor_a_ref)

        rank = require_positive_rank(
            require_numeric(payload, op_name=self.name, key="rank"),
            error_type=PhloraTransformError,
            op_name=self.name,
            key="rank",
        )
        delete_source = _require_boolean(
            payload,
            op_name=self.name,
            key="delete_original",
            default=True,
        )
        require_missing_outputs = _require_boolean(
            payload,
            op_name=self.name,
            key="require_missing_dest",
            default=True,
        )
        return PhloraSpec(
            source_ref=source_ref,
            factor_b_ref=factor_b_ref,
            factor_a_ref=factor_a_ref,
            rank=rank,
            delete_source=delete_source,
            require_missing_outputs=require_missing_outputs,
        )

    def validate_refs(self, from_a_ref: TensorRef, from_b_ref: TensorRef, to_ref: TensorRef) -> None:
        if from_a_ref.slice_spec is not None:
            raise PhloraTransformError("phlora target must not be sliced")
        if from_b_ref.slice_spec is not None:
            raise PhloraTransformError("phlora target_b must not be sliced")
        if to_ref.slice_spec is not None:
            raise PhloraTransformError("phlora target_a must not be sliced")

    def infer_output_model(self, spec: object) -> str:
        typed = self.require_spec(spec)
        model = typed.source_ref.model
        if model is None:
            raise PhloraTransformError("phlora output model missing")
        return model

    def resolve_items(
        self,
        spec: PhloraSpec,
        provider: StateDictProvider,
    ) -> list[ResolvedPhloraMapping]:
        factor_a_mappings = resolve_name_mappings(
            from_ref=spec.source_ref,
            to_ref=spec.factor_a_ref,
            provider=provider,
            op_name=self.name,
        )
        factor_b_mappings = resolve_name_mappings(
            from_ref=spec.source_ref,
            to_ref=spec.factor_b_ref,
            provider=provider,
            op_name=self.name,
        )
        pairs = _pair_mappings(factor_a_mappings, factor_b_mappings, op_name=self.name)

        resolved: list[ResolvedPhloraMapping] = []
        for source_item, factor_a_item, factor_b_item in pairs:
            resolved.append(
                ResolvedPhloraMapping(
                    source_model=source_item.src_model,
                    source_name=source_item.src_name,
                    factor_a_model=factor_a_item.dst_model,
                    factor_a_name=factor_a_item.dst_name,
                    factor_b_model=factor_b_item.dst_model,
                    factor_b_name=factor_b_item.dst_name,
                )
            )
        return resolved

    def apply_item(
        self,
        spec: PhloraSpec,
        item: ResolvedPhloraMapping,
        provider: StateDictProvider,
    ) -> None:
        src_sd = provider.get_state_dict(item.source_model)
        factor_a_sd = provider.get_state_dict(item.factor_a_model)
        factor_b_sd = provider.get_state_dict(item.factor_b_model)

        if item.source_name not in src_sd:
            raise PhloraTransformError(
                f"phlora source disappeared during apply: {item.source_model}::{item.source_name}"
            )

        source = src_sd[item.source_name]

        if spec.require_missing_outputs and item.factor_a_name in factor_a_sd:
            raise PhloraTransformError(
                f"phlora destination already exists: {item.factor_a_model}::{item.factor_a_name}"
            )
        if spec.require_missing_outputs and item.factor_b_name in factor_b_sd:
            raise PhloraTransformError(
                f"phlora destination already exists: {item.factor_b_model}::{item.factor_b_name}"
            )
        if item.factor_a_model == item.factor_b_model and item.factor_a_name == item.factor_b_name:
            raise PhloraTransformError(
                f"phlora destination collision for source {item.source_model}::{item.source_name}: "
                f"{item.factor_a_model}::{item.factor_a_name}"
            )

        lora_a, lora_b = compute_phlora_factors(
            source,
            spec.rank,
            cache=self._svd_cache,
            cache_key=f"{item.source_model}::{item.source_name}",
            error_type=PhloraTransformError,
            op_name="phlora",
            tensor_name=f"{item.source_model}::{item.source_name}",
        )

        factor_a_sd[item.factor_a_name] = lora_a.to(dtype=source.dtype, device=source.device)
        factor_b_sd[item.factor_b_name] = lora_b.to(dtype=source.dtype, device=source.device)

        if spec.delete_source:
            del src_sd[item.source_name]


def _pair_mappings(
    factor_a_mappings: list[ResolvedMapping],
    factor_b_mappings: list[ResolvedMapping],
    *,
    op_name: str,
) -> list[tuple[ResolvedMapping, ResolvedMapping, ResolvedMapping]]:
    factor_a_by_source = {(m.src_model, m.src_name): m for m in factor_a_mappings}
    factor_b_by_source = {(m.src_model, m.src_name): m for m in factor_b_mappings}
    if set(factor_a_by_source.keys()) != set(factor_b_by_source.keys()):
        raise PhloraTransformError(f"{op_name} target_a/target_b mappings do not match source set")
    pairs: list[tuple[ResolvedMapping, ResolvedMapping, ResolvedMapping]] = []
    for key in sorted(factor_a_by_source.keys()):
        source_item = factor_a_by_source[key]
        pairs.append((source_item, factor_a_by_source[key], factor_b_by_source[key]))
    return pairs


def _require_boolean(payload: dict, *, op_name: str, key: str, default: bool) -> bool:
    if key not in payload:
        return default
    value = payload[key]
    if not isinstance(value, bool):
        raise PhloraTransformError(f"{op_name}.{key} must be a boolean when provided")
    return value




register_transform(PhloraTransform())


class PhloraInPlaceTransformError(TransformError):
    pass


@dataclass(frozen=True)
class PhloraInPlaceSpec(UnarySpec):
    rank: int


class PhloraInPlaceTransform(UnaryTransform[PhloraInPlaceSpec]):
    name = "phlora_"
    error_type = PhloraInPlaceTransformError
    spec_type = PhloraInPlaceSpec
    allowed_keys = {"target", "rank"}
    required_keys = {"target", "rank"}
    progress_desc = "Applying phlora_ transforms"
    help_text = (
        "Applies in-place PHLoRA low-rank reconstruction.\n"
        "\n"
        "For each matched 2D tensor W:\n"
        "  u, s, vh = svd(W)\n"
        "  W <- (u[:, :r] * s[:r]) @ vh[:r, :]\n"
        "\n"
        "Examples:\n"
        "  phlora_: { target: '.*weight', rank: 64 }\n"
        "  phlora_: { target: h.0.attn.c_proj.weight, rank: 16 }"
    )

    def __init__(self) -> None:
        super().__init__()
        self._svd_cache = PhloraSvdCache()

    def build_spec(self, target_ref: TensorRef, payload: dict) -> PhloraInPlaceSpec:
        rank = require_positive_rank(
            require_numeric(payload, op_name=self.name, key="rank"),
            error_type=PhloraInPlaceTransformError,
            op_name=self.name,
            key="rank",
        )
        return PhloraInPlaceSpec(target_ref=target_ref, rank=rank)

    def apply_to_target(
        self,
        spec: PhloraInPlaceSpec,
        name: str,
        provider: StateDictProvider,
    ) -> None:
        model = must_model(spec.target_ref)
        sd = provider.get_state_dict(model)

        source = sd[name]
        new_tensor = reconstruct_phlora_rank(
            source,
            spec.rank,
            cache=self._svd_cache,
            cache_key=f"{model}::{name}",
            error_type=PhloraInPlaceTransformError,
            op_name="phlora_",
            tensor_name=f"{model}::{name}",
        )
        sd[name] = new_tensor.to(dtype=source.dtype, device=source.device)


register_transform(PhloraInPlaceTransform())



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import torch

from ..core import TransformError, collect_ref_models, compile_tensor_ref_expr, format_ref, register_assert_expr, resolve_tensors, require_mapping_assert_payload
from ..core import TensorRef
from ..core import StateDictProvider
from ..core import ScalarComparison, parse_scalar_comparison


@dataclass(frozen=True)
class DimensionsExpr:
    ref: TensorRef
    comparison: ScalarComparison

    def evaluate(self, provider: StateDictProvider) -> None:
        for ref, tensor in resolve_tensors(self.ref, provider, op_name="dimensions.of"):
            actual = len(tensor.shape)
            if not self.comparison.matches(actual):
                raise TransformError(
                    f"dimensions failed: {format_ref(ref)} has {actual} dims, expected {self.comparison.describe()}"
                )

    def collect_models(self) -> set[str]:
        return collect_ref_models(self.ref)


@register_assert_expr(
    "dimensions",
    payload_kind="mapping",
    allowed_keys={"of", "is", "ge", "gt", "le", "lt"},
    required_keys={"of"},
    description="Succeeds if the tensor rank satisfies the requested comparison(s).",
)
def compile_dimensions_expr(payload: Any, default_model: str | None) -> DimensionsExpr:
    payload = require_mapping_assert_payload(
        payload,
        op_name="dimensions",
        allowed_keys={"of", "is", "ge", "gt", "le", "lt"},
        required_keys={"of"},
    )
    ref = compile_tensor_ref_expr(payload["of"], default_model, "dimensions.of")
    comparison = parse_scalar_comparison(payload, op_name="dimensions")
    return DimensionsExpr(ref=ref, comparison=comparison)



from __future__ import annotations

from typing import Any, Callable

import torch

from .name_mapping import ResolvedMapping, require_dest_present, resolve_name_mappings
from .refs import TensorRef, format_tensor_ref, must_model, parse_slice, select_tensor
from .types import StateDictProvider, TransformError


def resolve_target_names(
    *,
    target_ref: TensorRef,
    provider: StateDictProvider,
    op_name: str,
    match_names: Callable[..., list[str]],
    error_type: type[TransformError],
) -> list[str]:
    model = must_model(target_ref)
    sd = provider.get_state_dict(model)

    try:
        matches = match_names(
            expr=target_ref.expr,
            names=sd.keys(),
            op_name=op_name,
            role="target",
        )
    except TransformError as exc:
        raise error_type(str(exc)) from exc

    if not matches:
        raise error_type(f"{op_name} matched zero tensors: {format_tensor_ref(target_ref)}")

    return matches


def resolve_single_tensor(
    ref: TensorRef,
    provider: StateDictProvider,
    *,
    op_name: str,
    resolve_names: Any,
    error_type: type[TransformError],
) -> torch.Tensor:
    model = must_model(ref)
    sd = provider.get_state_dict(model)
    matches = resolve_names(ref, provider, op_name=op_name)

    if len(matches) == 0:
        raise error_type(f"{op_name} failed: {format_tensor_ref(ref)} matched zero tensors")
    if len(matches) != 1:
        raise error_type(
            f"{op_name} failed: {format_tensor_ref(ref)} matched {len(matches)} tensors, expected 1"
        )

    tensor = sd[matches[0]]
    slice_spec = parse_slice(ref.slice_spec) if ref.slice_spec else None
    return select_tensor(tensor, slice_spec)


def _resolve_tensors(
    ref: TensorRef,
    provider: StateDictProvider,
    *,
    op_name: str,
    resolve_names: Any,
) -> list[tuple[TensorRef, torch.Tensor]]:
    model = must_model(ref)
    sd = provider.get_state_dict(model)
    matches = resolve_names(ref, provider, op_name=op_name)
    slice_spec = parse_slice(ref.slice_spec) if ref.slice_spec else None
    resolved: list[tuple[TensorRef, torch.Tensor]] = []

    for name in matches:
        resolved_ref = TensorRef(model=model, expr=name, slice_spec=ref.slice_spec)
        resolved.append((resolved_ref, select_tensor(sd[name], slice_spec)))

    return resolved


def _resolve_tensor_mappings(
    from_ref: TensorRef,
    to_ref: TensorRef,
    provider: StateDictProvider,
    *,
    op_name: str,
    error_type: type[TransformError],
) -> list[tuple[TensorRef, torch.Tensor, TensorRef, torch.Tensor]]:
    try:
        mappings = resolve_name_mappings(
            from_ref=from_ref,
            to_ref=to_ref,
            provider=provider,
            op_name=op_name,
        )
        require_dest_present(mappings=mappings, provider=provider, op_name=op_name)
    except TransformError as exc:
        raise error_type(str(exc)) from exc

    resolved: list[tuple[TensorRef, torch.Tensor, TensorRef, torch.Tensor]] = []
    for item in mappings:
        resolved.append(_resolve_mapping_tensors(item, from_ref=from_ref, to_ref=to_ref, provider=provider))
    return resolved


def _resolve_mapping_tensors(
    item: ResolvedMapping,
    *,
    from_ref: TensorRef,
    to_ref: TensorRef,
    provider: StateDictProvider,
) -> tuple[TensorRef, torch.Tensor, TensorRef, torch.Tensor]:
    src_sd = provider.get_state_dict(item.src_model)
    dst_sd = provider.get_state_dict(item.dst_model)

    left_ref = TensorRef(model=item.src_model, expr=item.src_name, slice_spec=from_ref.slice_spec)
    right_ref = TensorRef(model=item.dst_model, expr=item.dst_name, slice_spec=to_ref.slice_spec)
    left = select_tensor(src_sd[item.src_name], item.src_slice)
    right = select_tensor(dst_sd[item.dst_name], item.dst_slice)
    return left_ref, left, right_ref, right


__all__ = [
    "resolve_target_names",
    "resolve_single_tensor",
]



from importlib import import_module
from pkgutil import iter_modules


for module_info in iter_modules(__path__):  # type: ignore[name-defined]
    module_name = module_info.name
    if module_name.startswith("_"):
        continue
    import_module(f"{__name__}.{module_name}")



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
import re
from typing import Dict, Generic, Iterable, List, Literal, TypeVar

from tqdm import tqdm

from .name_mapping import (
    _match_structured_expr,
    _rewrite_structured_expr,
    ResolvedMapping,
    match_expr_names,
    require_dest_missing,
    require_dest_present,
    resolve_name_mappings,
)
from .refs import TensorRef, format_tensor_ref, must_model, parse_model_expr, parse_slice
from .resolver import resolve_target_names as resolve_target_names_generic
from .types import StateDictProvider, TransformError
from .validation import ensure_mapping_payload, require_expr, validate_payload_keys


class TransformControl(Enum):
    CONTINUE = "continue"
    EXIT = "exit"


@dataclass(frozen=True)
class TransformResult:
    name: str
    count: int
    control: TransformControl = TransformControl.CONTINUE


@dataclass(frozen=True)
class CompiledTransform:
    transform: "BaseTransform"
    spec: object


class BaseTransform(ABC):
    name: str
    completion_requires_payload: bool = True

    @abstractmethod
    def compile(self, payload: dict, default_model: str | None) -> object:
        raise NotImplementedError

    @abstractmethod
    def apply(self, spec: object, provider: StateDictProvider) -> TransformResult:
        raise NotImplementedError

    @abstractmethod
    def infer_output_model(self, spec: object) -> str:
        raise NotImplementedError

    def contributes_output_model(self, spec: object) -> bool:
        del spec
        return True

    def completion_reference_keys(self) -> list[str]:
        return []

    def completion_payload_start_candidates(self, prefix_text: str) -> list[str] | None:
        del prefix_text
        return None

    def completion_key_candidates(self, before_cursor: str, prefix_text: str) -> list[str] | None:
        del before_cursor, prefix_text
        return None

    def completion_value_candidates(
        self,
        value_key: str | None,
        prefix_text: str,
        model_aliases: list[str],
    ) -> list[str] | None:
        del value_key, prefix_text, model_aliases
        return None

    def completion_committed_next_candidates(self, value_key: str | None) -> list[str] | None:
        del value_key
        return None


SpecT = TypeVar("SpecT")


class TypedTransform(BaseTransform, ABC, Generic[SpecT]):
    error_type: type[TransformError] = TransformError
    spec_type: type[SpecT]

    def require_spec(self, spec: object) -> SpecT:
        if not isinstance(spec, self.spec_type):
            raise self.error_type(
                f"{self.name} expected {self.spec_type.__name__}, got {type(spec).__name__}"
            )
        return spec


REGISTRY: Dict[str, BaseTransform] = {}


def register_transform(transform: BaseTransform) -> None:
    name = getattr(transform, "name", None)
    if not isinstance(name, str) or not name:
        raise TransformError("transform must define a non-empty string 'name'")
    if name in REGISTRY:
        raise TransformError(f"transform already registered: {name}")
    REGISTRY[name] = transform


def get_transform(name: str) -> BaseTransform:
    try:
        return REGISTRY[name]
    except KeyError as exc:
        raise TransformError(f"unknown transform: {name}") from exc


def list_transforms() -> List[str]:
    return sorted(REGISTRY.keys())


def apply_transform(compiled: CompiledTransform, provider: StateDictProvider) -> TransformResult:
    return compiled.transform.apply(compiled.spec, provider)


class DestinationPolicy(Enum):
    ANY = "any"
    MUST_EXIST = "must_exist"
    MUST_NOT_EXIST = "must_not_exist"


ItemT = TypeVar("ItemT")


class IteratingTransform(TypedTransform[SpecT], ABC, Generic[SpecT, ItemT]):
    error_type: type[TransformError] = TransformError
    spec_type: type[SpecT]
    progress_desc: str | None = None
    progress_unit: str = "tensor"

    def apply(self, spec: object, provider: StateDictProvider) -> TransformResult:
        typed = self.require_spec(spec)
        items = self.resolve_items(typed, provider)
        self.validate_resolved_items(typed, items, provider)

        for item in self.iter_items(items):
            self.apply_item(typed, item, provider)

        return TransformResult(name=self.name, count=len(items))

    def iter_items(self, items: list[ItemT]) -> Iterable[ItemT]:
        if self.progress_desc is None:
            return items
        return tqdm(items, desc=self.progress_desc, unit=self.progress_unit)

    @abstractmethod
    def resolve_items(
        self,
        spec: SpecT,
        provider: StateDictProvider,
    ) -> list[ItemT]:
        ...

    def validate_resolved_items(
        self,
        spec: SpecT,
        items: list[ItemT],
        provider: StateDictProvider,
    ) -> None:
        del spec, items, provider

    @abstractmethod
    def apply_item(
        self,
        spec: SpecT,
        item: ItemT,
        provider: StateDictProvider,
    ) -> None:
        ...


@dataclass(frozen=True)
class UnarySpec:
    target_ref: TensorRef

    def collect_models(self) -> set[str]:
        return {must_model(self.target_ref)}


UnarySpecT = TypeVar("UnarySpecT", bound=UnarySpec)


def _resolve_unary_target_names(
    *,
    target_ref: TensorRef,
    provider: StateDictProvider,
    op_name: str,
    error_type: type[TransformError],
) -> list[str]:
    return resolve_target_names_generic(
        target_ref=target_ref,
        provider=provider,
        op_name=op_name,
        match_names=match_expr_names,
        error_type=error_type,
    )


class UnaryTransform(IteratingTransform[UnarySpecT, str], ABC, Generic[UnarySpecT]):
    error_type: type[TransformError] = TransformError
    spec_type: type[UnarySpecT]

    allowed_keys: set[str] = {"target"}
    required_keys: set[str] = {"target"}
    target_key: str = "target"
    slice_policy: Literal["allow", "forbid"] = "forbid"

    def completion_reference_keys(self) -> list[str]:
        return [self.target_key]

    def compile(self, payload: dict, default_model: str | None) -> UnarySpecT:
        payload = ensure_mapping_payload(payload, self.name)
        validate_payload_keys(
            payload,
            op_name=self.name,
            allowed_keys=self.allowed_keys,
            required_keys=self.required_keys,
        )

        raw_target = self.require_target_expr(payload)
        target_ref = parse_model_expr(raw_target, default_model=default_model)

        self.validate_target_ref(target_ref)

        assert target_ref.model is not None
        return self.build_spec(target_ref=target_ref, payload=payload)

    def infer_output_model(self, spec: object) -> str:
        typed = self.require_spec(spec)
        model = typed.target_ref.model
        if model is None:
            raise self.error_type(f"{self.name} output model missing")
        return model

    def require_target_expr(self, payload: dict) -> str | list[object]:
        return require_expr(payload, op_name=self.name, key=self.target_key)

    def build_spec(self, target_ref: TensorRef, payload: dict) -> UnarySpecT:
        return self.spec_type(target_ref=target_ref)

    def validate_target_ref(self, target_ref: TensorRef) -> None:
        if target_ref.slice_spec is None:
            return

        if self.slice_policy == "forbid":
            raise self.error_type(f"{self.name} target must not be sliced")

        parse_slice(target_ref.slice_spec)

    def resolve_items(self, spec: UnarySpecT, provider: StateDictProvider) -> list[str]:
        return _resolve_unary_target_names(
            target_ref=spec.target_ref,
            provider=provider,
            op_name=self.name,
            error_type=self.error_type,
        )

    def resolve_targets(self, spec: UnarySpecT, provider: StateDictProvider) -> list[str]:
        return self.resolve_items(spec, provider)

    @abstractmethod
    def apply_to_target(self, spec: UnarySpecT, name: str, provider: StateDictProvider) -> None:
        ...

    def apply_item(self, spec: UnarySpecT, item: str, provider: StateDictProvider) -> None:
        self.apply_to_target(spec, item, provider)


@dataclass(frozen=True)
class BinaryMappingSpec:
    from_ref: TensorRef
    to_ref: TensorRef

    def collect_models(self) -> set[str]:
        return {must_model(self.from_ref), must_model(self.to_ref)}


BinarySpecT = TypeVar("BinarySpecT", bound=BinaryMappingSpec)


class BinaryMappingTransform(IteratingTransform[BinarySpecT, ResolvedMapping], ABC, Generic[BinarySpecT]):
    error_type: type[TransformError] = TransformError
    spec_type: type[BinarySpecT]
    destination_policy: DestinationPolicy = DestinationPolicy.ANY

    allowed_keys = {"from", "to"}
    required_keys = {"from", "to"}

    def completion_reference_keys(self) -> list[str]:
        return ["from", "to"]

    def compile(self, payload: dict, default_model: str | None) -> BinarySpecT:
        payload = ensure_mapping_payload(payload, self.name)
        validate_payload_keys(
            payload,
            op_name=self.name,
            allowed_keys=self.allowed_keys,
            required_keys=self.required_keys,
        )

        from_ref, to_ref = self.parse_refs(payload, default_model)
        self.validate_refs(from_ref, to_ref)

        assert from_ref.model is not None
        assert to_ref.model is not None
        return self.build_spec(from_ref=from_ref, to_ref=to_ref)

    def infer_output_model(self, spec: object) -> str:
        typed = self.require_spec(spec)
        model = typed.to_ref.model
        if model is None:
            raise self.error_type(f"{self.name} output model missing")
        return model

    def parse_refs(
        self,
        payload: dict,
        default_model: str | None,
    ) -> tuple[TensorRef, TensorRef]:
        raw_from = require_expr(payload, op_name=self.name, key="from")
        raw_to = require_expr(payload, op_name=self.name, key="to")

        from_ref = parse_model_expr(raw_from, default_model=default_model)
        to_ref = parse_model_expr(raw_to, default_model=default_model)
        return from_ref, to_ref

    def build_spec(self, from_ref: TensorRef, to_ref: TensorRef) -> BinarySpecT:
        return self.spec_type(from_ref=from_ref, to_ref=to_ref)

    def resolve_items(
        self,
        spec: BinarySpecT,
        provider: StateDictProvider,
    ) -> list[ResolvedMapping]:
        return resolve_name_mappings(
            from_ref=spec.from_ref,
            to_ref=spec.to_ref,
            provider=provider,
            op_name=self.name,
        )

    @abstractmethod
    def validate_refs(self, from_ref: TensorRef, to_ref: TensorRef) -> None:
        ...

    def validate_resolved_items(
        self,
        spec: BinarySpecT,
        mappings: list[ResolvedMapping],
        provider: StateDictProvider,
    ) -> None:
        del spec
        if self.destination_policy is DestinationPolicy.MUST_EXIST:
            require_dest_present(
                mappings=mappings,
                provider=provider,
                op_name=self.name,
            )
            return

        if self.destination_policy is DestinationPolicy.MUST_NOT_EXIST:
            require_dest_missing(
                mappings=mappings,
                provider=provider,
                op_name=self.name,
            )

    def apply_mapping(self, item: ResolvedMapping, provider: StateDictProvider) -> None:
        del item, provider
        raise NotImplementedError

    def apply_item(self, spec: BinarySpecT, item: ResolvedMapping, provider: StateDictProvider) -> None:
        del spec
        self.apply_mapping(item, provider)


@dataclass(frozen=True)
class TernaryMappingSpec:
    from_a_ref: TensorRef
    from_b_ref: TensorRef
    to_ref: TensorRef

    def collect_models(self) -> set[str]:
        return {
            must_model(self.from_a_ref),
            must_model(self.from_b_ref),
            must_model(self.to_ref),
        }


@dataclass(frozen=True)
class ResolvedTernaryMapping:
    src_a_model: str
    src_a_name: str
    src_a_slice: tuple[object, ...] | None
    src_b_model: str
    src_b_name: str
    src_b_slice: tuple[object, ...] | None
    dst_model: str
    dst_name: str
    dst_slice: tuple[object, ...] | None


TernarySpecT = TypeVar("TernarySpecT", bound=TernaryMappingSpec)


class TernaryMappingTransform(IteratingTransform[TernarySpecT, ResolvedTernaryMapping], ABC, Generic[TernarySpecT]):
    error_type: type[TransformError] = TransformError
    spec_type: type[TernarySpecT]
    destination_policy: DestinationPolicy = DestinationPolicy.ANY

    allowed_keys = {"from_a", "from_b", "to"}
    required_keys = {"from_a", "from_b", "to"}

    def completion_reference_keys(self) -> list[str]:
        return ["from_a", "from_b", "to"]

    def compile(self, payload: dict, default_model: str | None) -> TernarySpecT:
        payload = ensure_mapping_payload(payload, self.name)
        validate_payload_keys(
            payload,
            op_name=self.name,
            allowed_keys=self.allowed_keys,
            required_keys=self.required_keys,
        )

        from_a_ref, from_b_ref, to_ref = self.parse_refs(payload, default_model)
        self.validate_refs(from_a_ref, from_b_ref, to_ref)

        assert from_a_ref.model is not None
        assert from_b_ref.model is not None
        assert to_ref.model is not None
        return self.build_spec(from_a_ref=from_a_ref, from_b_ref=from_b_ref, to_ref=to_ref)

    def infer_output_model(self, spec: object) -> str:
        typed = self.require_spec(spec)
        model = typed.to_ref.model
        if model is None:
            raise self.error_type(f"{self.name} output model missing")
        return model

    def parse_refs(
        self,
        payload: dict,
        default_model: str | None,
    ) -> tuple[TensorRef, TensorRef, TensorRef]:
        raw_from_a = require_expr(payload, op_name=self.name, key="from_a")
        raw_from_b = require_expr(payload, op_name=self.name, key="from_b")
        raw_to = require_expr(payload, op_name=self.name, key="to")

        from_a_ref = parse_model_expr(raw_from_a, default_model=default_model)
        from_b_ref = parse_model_expr(raw_from_b, default_model=default_model)
        to_ref = parse_model_expr(raw_to, default_model=default_model)
        return from_a_ref, from_b_ref, to_ref

    def build_spec(self, from_a_ref: TensorRef, from_b_ref: TensorRef, to_ref: TensorRef) -> TernarySpecT:
        return self.spec_type(from_a_ref=from_a_ref, from_b_ref=from_b_ref, to_ref=to_ref)

    @abstractmethod
    def validate_refs(self, from_a_ref: TensorRef, from_b_ref: TensorRef, to_ref: TensorRef) -> None:
        ...

    def resolve_items(
        self,
        spec: TernarySpecT,
        provider: StateDictProvider,
    ) -> list[ResolvedTernaryMapping]:
        from_a_ref = spec.from_a_ref
        from_b_ref = spec.from_b_ref
        to_ref = spec.to_ref

        a_slice = parse_slice(from_a_ref.slice_spec) if from_a_ref.slice_spec is not None else None
        b_slice = parse_slice(from_b_ref.slice_spec) if from_b_ref.slice_spec is not None else None
        dst_slice = parse_slice(to_ref.slice_spec) if to_ref.slice_spec is not None else None

        src_a_model = must_model(from_a_ref)
        src_b_model = must_model(from_b_ref)
        dst_model = must_model(to_ref)
        src_a_sd = provider.get_state_dict(src_a_model)
        src_b_sd = provider.get_state_dict(src_b_model)

        a_expr = from_a_ref.expr
        b_expr = from_b_ref.expr
        to_expr = to_ref.expr
        if isinstance(a_expr, str) and isinstance(b_expr, str) and isinstance(to_expr, str):
            src_a_names = match_expr_names(
                expr=a_expr,
                names=src_a_sd.keys(),
                op_name=self.name,
                role="source_a",
            )
            if not src_a_names:
                raise self.error_type(
                    f"{self.name} source_a matched zero tensors: "
                    f"{format_tensor_ref(from_a_ref)}; available tensors: {sorted(src_a_sd.keys())}"
                )

            resolved: list[ResolvedTernaryMapping] = []
            dst_names_seen: set[str] = set()
            for src_a_name in src_a_names:
                try:
                    src_b_name = re.sub(a_expr, b_expr, src_a_name)
                    dst_name = re.sub(a_expr, to_expr, src_a_name)
                except re.error as exc:
                    raise self.error_type(
                        f"{self.name} invalid regex rewrite from {a_expr!r}: {exc}"
                    ) from exc

                if src_b_name not in src_b_sd:
                    raise self.error_type(
                        f"{self.name} source_b missing: {src_b_model}::{src_b_name}"
                    )
                if dst_name in dst_names_seen:
                    raise self.error_type(f"{self.name} destination collision: {dst_model}::{dst_name}")
                dst_names_seen.add(dst_name)

                resolved.append(
                    ResolvedTernaryMapping(
                        src_a_model=src_a_model,
                        src_a_name=src_a_name,
                        src_a_slice=a_slice,
                        src_b_model=src_b_model,
                        src_b_name=src_b_name,
                        src_b_slice=b_slice,
                        dst_model=dst_model,
                        dst_name=dst_name,
                        dst_slice=dst_slice,
                    )
                )
            return resolved

        if isinstance(a_expr, list) and isinstance(b_expr, list) and isinstance(to_expr, list):
            resolved: list[ResolvedTernaryMapping] = []
            dst_names_seen: set[str] = set()
            matched_any = False
            for src_a_name in sorted(src_a_sd.keys()):
                match = _match_structured_expr(
                    expr=a_expr,
                    name=src_a_name,
                    op_name=self.name,
                    role="source_a",
                )
                if match is None:
                    continue
                matched_any = True

                src_b_name = _rewrite_structured_expr(
                    expr=b_expr,
                    match=match,
                    op_name=self.name,
                    role="source_b",
                )
                dst_name = _rewrite_structured_expr(
                    expr=to_expr,
                    match=match,
                    op_name=self.name,
                    role="destination",
                )

                if src_b_name not in src_b_sd:
                    raise self.error_type(
                        f"{self.name} source_b missing: {src_b_model}::{src_b_name}"
                    )
                if dst_name in dst_names_seen:
                    raise self.error_type(f"{self.name} destination collision: {dst_model}::{dst_name}")
                dst_names_seen.add(dst_name)

                resolved.append(
                    ResolvedTernaryMapping(
                        src_a_model=src_a_model,
                        src_a_name=src_a_name,
                        src_a_slice=a_slice,
                        src_b_model=src_b_model,
                        src_b_name=src_b_name,
                        src_b_slice=b_slice,
                        dst_model=dst_model,
                        dst_name=dst_name,
                        dst_slice=dst_slice,
                    )
                )

            if not matched_any:
                raise self.error_type(
                    f"{self.name} source_a matched zero tensors: "
                    f"{format_tensor_ref(from_a_ref)}; available tensors: {sorted(src_a_sd.keys())}"
                )
            return resolved

        raise self.error_type(
            f"{self.name} requires from_a/from_b/to expressions of the same kind: "
            "either all strings (regex mode) or all lists (structured mode)"
        )

    def validate_resolved_items(
        self,
        spec: TernarySpecT,
        mappings: list[ResolvedTernaryMapping],
        provider: StateDictProvider,
    ) -> None:
        del spec
        if self.destination_policy is DestinationPolicy.ANY:
            return

        for item in mappings:
            dst_sd = provider.get_state_dict(item.dst_model)
            exists = item.dst_name in dst_sd

            if self.destination_policy is DestinationPolicy.MUST_EXIST and not exists:
                raise self.error_type(
                    f"{self.name} destination missing: {item.dst_model}::{item.dst_name}"
                )
            if self.destination_policy is DestinationPolicy.MUST_NOT_EXIST and exists:
                raise self.error_type(
                    f"{self.name} destination already exists: {item.dst_model}::{item.dst_name}"
                )

    @abstractmethod
    def apply_mapping(self, item: ResolvedTernaryMapping, provider: StateDictProvider) -> None:
        ...

    def apply_item(
        self,
        spec: TernarySpecT,
        item: ResolvedTernaryMapping,
        provider: StateDictProvider,
    ) -> None:
        del spec
        self.apply_mapping(item, provider)


__all__ = [
    "BaseTransform",
    "BinaryMappingSpec",
    "BinaryMappingTransform",
    "CompiledTransform",
    "DestinationPolicy",
    "IteratingTransform",
    "ResolvedTernaryMapping",
    "TernaryMappingSpec",
    "TernaryMappingTransform",
    "TransformControl",
    "TransformResult",
    "TypedTransform",
    "UnarySpec",
    "UnaryTransform",
    "REGISTRY",
    "apply_transform",
    "get_transform",
    "list_transforms",
    "register_transform",
]



from __future__ import annotations

from dataclasses import dataclass
import re
from typing import Any, Literal

import torch

from ..core import match_expr_names
from ..core import TensorRef, format_tensor_ref, must_model, parse_model_expr, parse_slice, select_tensor
from ..core import TransformError
from ..core import TypedTransform, TransformResult, register_transform
from ..core import ensure_mapping_payload, require_nonempty_string, validate_payload_keys
from ..core import StateDictProvider
from ..engine.frontend import emit_line


class DiffTransformError(TransformError):
    pass


DiffMode = Literal["refs", "aliases"]


@dataclass(frozen=True)
class DiffSpec:
    left_ref: TensorRef
    right_ref: TensorRef
    eps: float | None
    mode: DiffMode

    def collect_models(self) -> set[str]:
        return {must_model(self.left_ref), must_model(self.right_ref)}


class DiffTransform(TypedTransform[DiffSpec]):
    name = "diff"
    error_type = DiffTransformError
    spec_type = DiffSpec
    allowed_keys = {"mode", "left", "right", "left_alias", "right_alias", "eps"}
    help_text = (
        "Compares two tensor sets and reports missing names on both sides plus tensors "
        "whose contents differ.\n"
        "\n"
        "Modes:\n"
        "  - refs (default): compare tensors selected by 'left' and 'right'\n"
        "  - aliases: compare all tensors from 'left_alias' and 'right_alias'\n"
        "\n"
        "Comparison is symmetric by tensor name. Shared names are compared by shape, dtype, "
        "device, then value equality. When 'eps' is provided, value comparisons use absolute "
        "tolerance.\n"
        "\n"
        "Examples:\n"
        "  diff: { left: base::'.*', right: edited::'.*' }\n"
        "  diff: { left: base::ln_f\\..*, right: edited::ln_f\\..*, eps: 1e-6 }\n"
        "  diff: { mode: aliases, left_alias: base, right_alias: edited }"
    )

    def compile(self, payload: Any, default_model: str | None) -> DiffSpec:
        payload = ensure_mapping_payload(payload, self.name)
        validate_payload_keys(
            payload,
            op_name=self.name,
            allowed_keys=self.allowed_keys,
        )

        raw_mode = payload.get("mode", "refs")
        if not isinstance(raw_mode, str) or not raw_mode:
            raise DiffTransformError("diff.mode must be a non-empty string when provided")
        mode = raw_mode.strip().lower()
        if mode not in {"refs", "aliases"}:
            raise DiffTransformError("diff.mode must be one of: refs, aliases")

        eps = _compile_eps(payload.get("eps"))

        if mode == "refs":
            _require_only_keys(payload, allowed={"mode", "left", "right", "eps"})
            if "left" not in payload:
                raise DiffTransformError("diff.left is required")
            if "right" not in payload:
                raise DiffTransformError("diff.right is required")

            left_ref = parse_model_expr(payload["left"], default_model=default_model)
            right_ref = parse_model_expr(payload["right"], default_model=default_model)
            _validate_slice(left_ref)
            _validate_slice(right_ref)
            return DiffSpec(left_ref=left_ref, right_ref=right_ref, eps=eps, mode="refs")

        _require_only_keys(payload, allowed={"mode", "left_alias", "right_alias", "eps"})
        left_alias = require_nonempty_string(payload, op_name=self.name, key="left_alias")
        right_alias = require_nonempty_string(payload, op_name=self.name, key="right_alias")
        return DiffSpec(
            left_ref=TensorRef(model=left_alias, expr=".*"),
            right_ref=TensorRef(model=right_alias, expr=".*"),
            eps=eps,
            mode="aliases",
        )

    def apply(self, spec: object, provider: StateDictProvider) -> TransformResult:
        typed = self.require_spec(spec)

        left_names = set(_resolve_names(typed.left_ref, provider))
        right_names = set(_resolve_names(typed.right_ref, provider))

        missing_on_left = sorted(right_names - left_names)
        missing_on_right = sorted(left_names - right_names)
        shared = sorted(left_names & right_names)
        differing = _collect_differences(shared, typed, provider)

        emit_line(
            f"Diff: {format_tensor_ref(typed.left_ref)} <-> {format_tensor_ref(typed.right_ref)}"
        )
        _echo_names("Missing on left", missing_on_left)
        _echo_names("Missing on right", missing_on_right)
        _echo_differences(differing)

        findings = len(missing_on_left) + len(missing_on_right) + len(differing)
        if findings == 0:
            emit_line("No differences found.")

        return TransformResult(name=self.name, count=findings)

    def infer_output_model(self, spec: object) -> str:
        del spec
        raise DiffTransformError("diff does not infer an output model")

    def contributes_output_model(self, spec: object) -> bool:
        del spec
        return False

    def completion_reference_keys(self) -> list[str]:
        return ["left", "right"]

    def completion_key_candidates(self, before_cursor: str, prefix_text: str) -> list[str] | None:
        used_keys = set(re.findall(r"([A-Za-z_][A-Za-z0-9_]*)\s*:", before_cursor))
        mode = _diff_mode(before_cursor)
        if mode == "aliases":
            options = ["mode: ", "left_alias: ", "right_alias: ", "eps: "]
        else:
            options = ["mode: ", "left: ", "right: ", "eps: "]
        filtered = [
            candidate
            for candidate in options
            if candidate[:-2] not in used_keys and candidate.startswith(prefix_text)
        ]
        if filtered:
            return filtered
        if options and not all(option[:-2] in used_keys for option in options):
            return []
        return ["}"]

    def completion_value_candidates(
        self,
        value_key: str | None,
        prefix_text: str,
        model_aliases: list[str],
    ) -> list[str] | None:
        if value_key == "mode":
            return [mode for mode in ("refs", "aliases") if mode.startswith(prefix_text)]
        if value_key in {"left_alias", "right_alias"}:
            return [alias for alias in model_aliases if alias.startswith(prefix_text)]
        return None


def _compile_eps(raw_eps: Any) -> float | None:
    if raw_eps is None:
        return None
    if isinstance(raw_eps, bool) or not isinstance(raw_eps, (int, float)):
        raise DiffTransformError("diff.eps must be a non-negative number")
    eps = float(raw_eps)
    if eps < 0:
        raise DiffTransformError("diff.eps must be a non-negative number")
    return eps


def _require_only_keys(payload: dict[str, object], *, allowed: set[str]) -> None:
    unexpected = set(payload) - allowed
    if unexpected:
        raise DiffTransformError(f"diff received unknown keys for this mode: {sorted(unexpected)}")


def _validate_slice(ref: TensorRef) -> None:
    if ref.slice_spec is not None:
        parse_slice(ref.slice_spec)


def _resolve_names(ref: TensorRef, provider: StateDictProvider) -> list[str]:
    model = must_model(ref)
    state_dict = provider.get_state_dict(model)
    try:
        return match_expr_names(
            expr=ref.expr,
            names=state_dict.keys(),
            op_name="diff",
            role="target",
        )
    except TransformError as exc:
        raise DiffTransformError(str(exc)) from exc


def _collect_differences(
    shared: list[str],
    spec: DiffSpec,
    provider: StateDictProvider,
) -> list[tuple[str, str]]:
    left_sd = provider.get_state_dict(must_model(spec.left_ref))
    right_sd = provider.get_state_dict(must_model(spec.right_ref))
    left_slice = parse_slice(spec.left_ref.slice_spec) if spec.left_ref.slice_spec else None
    right_slice = parse_slice(spec.right_ref.slice_spec) if spec.right_ref.slice_spec else None

    differing: list[tuple[str, str]] = []
    for name in shared:
        left = select_tensor(left_sd[name], left_slice)
        right = select_tensor(right_sd[name], right_slice)
        reason = _difference_reason(left, right, eps=spec.eps)
        if reason is not None:
            differing.append((name, reason))
    return differing


def _difference_reason(left: torch.Tensor, right: torch.Tensor, *, eps: float | None) -> str | None:
    if left.shape != right.shape:
        return f"shape {tuple(left.shape)} != {tuple(right.shape)}"
    if left.dtype != right.dtype:
        return f"dtype {left.dtype} != {right.dtype}"
    if left.device != right.device:
        return f"device {left.device} != {right.device}"

    if eps is None:
        if torch.equal(left, right):
            return None
        return "values differ"

    if left.is_complex():
        diff = torch.abs(left.to(torch.complex128) - right.to(torch.complex128))
    else:
        diff = torch.abs(left.to(torch.float64) - right.to(torch.float64))

    if bool(torch.all(diff <= eps).item()):
        return None

    max_abs_diff = float(diff.max().item())
    return f"values differ (max_abs_diff={max_abs_diff:g}, eps={eps:g})"


def _echo_names(title: str, names: list[str]) -> None:
    emit_line(f"{title}:")
    if not names:
        emit_line("  (none)")
        return
    for name in names:
        emit_line(f"  - {name}")


def _echo_differences(differing: list[tuple[str, str]]) -> None:
    emit_line("Differing:")
    if not differing:
        emit_line("  (none)")
        return
    for name, reason in differing:
        emit_line(f"  - {name}: {reason}")


def _diff_mode(before_cursor: str) -> str | None:
    match = re.search(r"\bmode\s*:\s*([A-Za-z_][A-Za-z0-9_]*)", before_cursor)
    if match is None:
        return None
    return match.group(1).lower()


register_transform(DiffTransform())



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Any

from ..engine import render_tree, summarize_tensor
from ..core import UnarySpec, UnaryTransform
from ..engine.checkpoint_io import tqdm
from ..engine.frontend import emit_line
from ..core import TensorRef, must_model, parse_model_expr, parse_slice, select_tensor
from ..core import TransformError
from ..core import TransformResult, register_transform
from ..core import ensure_mapping_payload, validate_payload_keys
from ..core import StateDictProvider


class DumpTransformError(TransformError):
    pass


@dataclass(frozen=True)
class DumpSpec(UnarySpec):
    format: str
    verbosity: str


class DumpTransform(UnaryTransform[DumpSpec]):
    name = "dump"
    error_type = DumpTransformError
    spec_type = DumpSpec
    allowed_keys = {"target", "format", "verbosity"}
    required_keys = set()
    slice_policy = "allow"
    progress_desc = "Dumping tensors"
    help_text = (
        "Displays tensors selected by 'target' without modifying them.\n"
        "\n"
        "Targets may be specified by name or pattern. Slices are written after '::', "
        "for example 'ln_f.weight::[:8]'. 'format' controls layout: 'tree' (default), "
        "'compact', or 'json'. 'verbosity' controls content: 'shape', 'stat' (default), "
        "or 'full'.\n"
        "\n"
        "Examples:\n"
        "  dump: { target: ln_f.weight }\n"
        "  dump: { target: '.*weight', format: compact, verbosity: shape }\n"
        "  dump: { target: 'h.0.attn.c_attn.weight::[:, :10]', format: json, verbosity: stat }\n"
        "  dump: { target: 'ln_f.weight::[:8]', format: tree, verbosity: full }"
    )

    def compile(self, payload: dict, default_model: str | None) -> DumpSpec:
        payload = ensure_mapping_payload(payload, self.name)
        validate_payload_keys(
            payload,
            op_name=self.name,
            allowed_keys=self.allowed_keys,
            required_keys=self.required_keys,
        )

        if "target" not in payload:
            payload = dict(payload)
            payload["target"] = ".*"

        raw_target = self.require_target_expr(payload)
        target_ref = parse_model_expr(raw_target, default_model=default_model)

        self.validate_target_ref(target_ref)

        assert target_ref.model is not None
        return self.build_spec(target_ref=target_ref, payload=payload)

    def build_spec(self, target_ref: TensorRef, payload: dict) -> DumpSpec:
        raw_format = payload.get("format", "compact")
        if not isinstance(raw_format, str) or not raw_format:
            raise DumpTransformError("dump.format must be a non-empty string")

        fmt = raw_format.strip().lower()
        if fmt not in {"json", "tree", "compact"}:
            raise DumpTransformError("dump.format must be one of: json, tree, compact")

        raw_verbosity = payload.get("verbosity", "shape")
        if not isinstance(raw_verbosity, str) or not raw_verbosity:
            raise DumpTransformError("dump.verbosity must be a non-empty string")

        verbosity = raw_verbosity.strip().lower()
        if verbosity not in {"shape", "stat", "full"}:
            raise DumpTransformError("dump.verbosity must be one of: shape, stat, full")

        return DumpSpec(target_ref=target_ref, format=fmt, verbosity=verbosity)

    def apply_to_target(self, spec: DumpSpec, name: str, provider: StateDictProvider) -> None:
        raise AssertionError("DumpTransform overrides apply() and does not use apply_to_target()")

    def contributes_output_model(self, spec: object) -> bool:
        del spec
        return False

    def apply(self, spec: object, provider: StateDictProvider) -> TransformResult:
        typed = self.require_spec(spec)

        model = must_model(typed.target_ref)
        sd = provider.get_state_dict(model)
        targets = self.resolve_targets(typed, provider)

        slice_spec = (
            parse_slice(typed.target_ref.slice_spec)
            if typed.target_ref.slice_spec is not None
            else None
        )

        tree: dict[str, Any] = {}

        for name in tqdm(targets, desc=self.progress_desc, unit="tensor"):
            tensor = sd[name]
            view = select_tensor(tensor, slice_spec)
            access_counts = _maybe_get_access_counts(sd, name, verbosity=typed.verbosity)
            insert_into_tree(
                tree,
                name.split("."),
                summarize_tensor(view, verbosity=typed.verbosity, access_counts=access_counts),
            )

        if typed.format == "json":
            emit_line(json.dumps(tree, separators=(",", ":"), sort_keys=True))
        elif typed.format == "tree":
            emit_line(render_tree(tree, compact=False))
        else:
            emit_line(render_tree(tree, compact=True))

        return TransformResult(name=self.name, count=len(targets))


def insert_into_tree(tree: dict[str, Any], parts: list[str], leaf: Any) -> None:
    node: Any = tree

    for i, part in enumerate(parts):
        is_last = i == len(parts) - 1
        next_is_index = i + 1 < len(parts) and parts[i + 1].isdigit()

        if part.isdigit():
            idx = int(part)

            if not isinstance(node, list):
                raise DumpTransformError("invalid tree structure while building dump")

            while len(node) <= idx:
                node.append(None)

            if is_last:
                node[idx] = leaf
                return

            child = node[idx]
            if child is None:
                child = [] if next_is_index else {}
                node[idx] = child
            elif not isinstance(child, (dict, list)):
                raise DumpTransformError("invalid tree structure while building dump")

            node = child
            continue

        if not isinstance(node, dict):
            raise DumpTransformError("invalid tree structure while building dump")

        if is_last:
            node[part] = leaf
            return

        child = node.get(part)
        if child is None:
            child = [] if next_is_index else {}
            node[part] = child
        elif not isinstance(child, (dict, list)):
            raise DumpTransformError("invalid tree structure while building dump")

        node = child


def _maybe_get_access_counts(
    state_dict: Any,
    key: str,
    *,
    verbosity: str,
) -> dict[str, int] | None:
    if verbosity not in {"stat", "full"}:
        return None
    access_counts = getattr(state_dict, "access_counts", None)
    if not callable(access_counts):
        return None
    return access_counts(key)












register_transform(DumpTransform())



from __future__ import annotations

import json
import logging
from contextlib import nullcontext
from pathlib import Path
from threading import Lock
from typing import Any, Dict, Literal, TypeVar

import torch
from safetensors.torch import load_file as load_safetensors_file
from safetensors.torch import save_file as save_safetensors_file

from ..core import StateDictLike
from .output_paths import resolve_sharded_output_directory
from .workers import choose_num_io_workers, run_threadpool_tasks_with_progress


try:
    from tqdm import tqdm
except ImportError:  # pragma: no cover
    T = TypeVar("T")

    class _TqdmDummy:
        def __init__(self, iterable=None, total=None, **_):
            self.iterable = iterable

        def __iter__(self):
            if self.iterable is None:
                return iter(())
            return iter(self.iterable)

        def update(self, *_):
            pass

        def close(self):
            pass

    def tqdm(iterable=None, **kwargs):  # type: ignore
        return _TqdmDummy(iterable, **kwargs)


logger = logging.getLogger("brainsurgery")


def tensor_nbytes(tensor: torch.Tensor) -> int:
    return tensor.numel() * tensor.element_size()


def shard_state_dict(
    state_dict: Dict[str, torch.Tensor],
    max_shard_size: int,
) -> list[dict[str, torch.Tensor]]:
    if max_shard_size <= 0:
        raise RuntimeError("max_shard_size must be positive")

    shards: list[dict[str, torch.Tensor]] = []
    current_shard: dict[str, torch.Tensor] = {}
    current_size = 0

    for key, tensor in state_dict.items():
        size = tensor_nbytes(tensor)

        if size > max_shard_size:
            if current_shard:
                shards.append(current_shard)
                current_shard = {}
                current_size = 0
            shards.append({key: tensor})
            continue

        if current_shard and current_size + size > max_shard_size:
            shards.append(current_shard)
            current_shard = {}
            current_size = 0

        current_shard[key] = tensor
        current_size += size

    if current_shard:
        shards.append(current_shard)

    return shards


def save_sharded_safetensors(
    state_dict: Dict[str, torch.Tensor],
    output_dir: Path,
    max_shard_size: int,
    *,
    max_io_workers: int,
) -> Path:
    output_dir.mkdir(parents=True, exist_ok=True)

    shards = shard_state_dict(state_dict, max_shard_size)
    total_size = sum(tensor_nbytes(tensor) for tensor in state_dict.values())
    total_shards = len(shards)

    logger.info(
        "Dividing preserved brain into %d safetensor shard(s) with maximum segment size %d bytes",
        total_shards,
        max_shard_size,
    )

    shard_infos: list[tuple[int, str, Path, dict[str, torch.Tensor]]] = []
    weight_map: dict[str, str] = {}

    for shard_index, shard in enumerate(shards, start=1):
        shard_name = f"model-{shard_index:05d}-of-{total_shards:05d}.safetensors"
        shard_path = output_dir / shard_name
        shard_infos.append((shard_index, shard_name, shard_path, shard))
        for key in shard:
            weight_map[key] = shard_name
    if len(weight_map) != len(state_dict):
        raise RuntimeError(
            "internal error while writing sharded safetensors: shard index coverage mismatch"
        )

    num_workers = choose_num_io_workers(total_shards, max_io_workers=max_io_workers)
    logger.info("Dispatching %d orderly thread(s) for preservation", num_workers)

    def _save_one_shard(
        item: tuple[int, str, Path, dict[str, torch.Tensor]],
    ) -> tuple[int, str]:
        shard_index, shard_name, shard_path, shard = item
        _save_safetensors_shard(shard_path, shard)
        return shard_index, shard_name

    def _on_shard_saved(
        item: tuple[int, str, Path, dict[str, torch.Tensor]],
        result: tuple[int, str],
    ) -> None:
        del item
        shard_index, shard_name = result
        logger.debug("Preserved shard %d/%d at %s", shard_index, total_shards, shard_name)

    run_threadpool_tasks_with_progress(
        items=shard_infos,
        worker=_save_one_shard,
        num_workers=num_workers,
        total=total_shards,
        progress_desc="Shard save",
        progress_unit="shard",
        progress_factory=tqdm,
        on_result=_on_shard_saved,
    )

    index = {
        "metadata": {"total_size": total_size},
        "weight_map": weight_map,
    }
    index_path = output_dir / "model.safetensors.index.json"
    index_path.write_text(json.dumps(index, indent=2), encoding="utf-8")
    return index_path


def _save_safetensors_shard(path: Path, shard: dict[str, torch.Tensor]) -> None:
    save_safetensors_file(shard, str(path))


def save_state_dict_to_path(
    state_dict: Dict[str, torch.Tensor],
    path: Path,
    *,
    format: Literal["safetensors", "torch"],
) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    if format == "torch":
        torch.save(state_dict, path)
        return
    save_safetensors_file(state_dict, str(path))


def persist_state_dict(
    state_dict: Dict[str, torch.Tensor],
    *,
    output_path: Path,
    output_format: Literal["safetensors", "torch"],
    shard_size: int | None,
    sharded_output_root: Path,
    max_io_workers: int,
) -> Path:
    if output_format == "torch":
        save_state_dict_to_path(state_dict, output_path, format="torch")
        return output_path

    if shard_size is None:
        save_state_dict_to_path(state_dict, output_path, format="safetensors")
        return output_path

    output_dir = resolve_sharded_output_directory(sharded_output_root, output_path)
    return save_sharded_safetensors(
        state_dict,
        output_dir,
        shard_size,
        max_io_workers=max_io_workers,
    )


def load_state_dict_from_path(path: Path, global_state_dict: StateDictLike, *, max_io_workers: int) -> None:
    if not path.exists():
        raise RuntimeError(f"checkpoint path does not exist: {path}")
    if path.is_dir():
        logger.info("CT scan shows a model directory at %s", path)
        return load_state_dict_from_directory(path, global_state_dict, max_io_workers=max_io_workers)
    logger.info("CT scan shows a single checkpoint file at %s", path)
    return load_state_dict_from_file(path, global_state_dict)


def load_state_dict_from_directory(path: Path, global_state_dict: StateDictLike, *, max_io_workers: int) -> None:
    pt_files = sorted(path.glob("*.pt")) + sorted(path.glob("*.pth")) + sorted(path.glob("*.bin"))
    safetensor_files = sorted(path.glob("*.safetensors"))
    index_file = path / "model.safetensors.index.json"

    if pt_files and safetensor_files:
        raise RuntimeError(
            f"model directory contains both torch and safetensors files; refusing ambiguous load: {path}"
        )

    if safetensor_files:
        if index_file.exists():
            logger.info("Detected safetensors index at %s", index_file)
            files = resolve_safetensor_shards_from_index(index_file, path)
        else:
            logger.info("No safetensors index found; exposing all safetensors shards")
            files = safetensor_files
    else:
        files = pt_files

    if not files:
        raise RuntimeError(f"no supported checkpoint files found in model directory: {path}")

    logger.info("Located %d checkpoint shard(s) in %s", len(files), path)

    num_workers = choose_num_io_workers(len(files), max_io_workers=max_io_workers)
    logger.info("Dispatching %d orderly thread(s) for exposure", num_workers)

    merge_lock = Lock()
    initial_count = len(global_state_dict)
    loaded_counts: list[int] = []

    def _load_one_file(file_path: Path) -> int:
        return load_state_dict_from_file(file_path, global_state_dict, merge_lock)

    def _on_file_loaded(file_path: Path, loaded_count: int) -> None:
        del file_path
        loaded_counts.append(loaded_count)

    run_threadpool_tasks_with_progress(
        items=files,
        worker=_load_one_file,
        num_workers=num_workers,
        total=len(files),
        progress_desc=f"Open {path.name}",
        progress_unit="file",
        progress_factory=tqdm,
        on_result=_on_file_loaded,
    )

    if len(loaded_counts) != len(files):
        raise RuntimeError(
            "internal error while loading checkpoint shards: progress count mismatch"
        )
    if any(count <= 0 for count in loaded_counts):
        raise RuntimeError("checkpoint shard file contained zero tensors")
    loaded_total = sum(loaded_counts)
    final_count = len(global_state_dict)
    if final_count - initial_count != loaded_total:
        raise RuntimeError(
            "internal error while loading checkpoint shards: merged tensor count mismatch"
        )

    logger.info("Cranial assembly complete for %s: %d tensor(s)", path, len(global_state_dict))


def resolve_safetensor_shards_from_index(index_file: Path, base_dir: Path) -> list[Path]:
    try:
        index_data = json.loads(index_file.read_text(encoding="utf-8"))
    except Exception as exc:
        raise RuntimeError(f"failed to parse safetensors index: {index_file}") from exc

    weight_map = index_data.get("weight_map")
    if not isinstance(weight_map, dict):
        raise RuntimeError(f"invalid safetensors index: missing weight_map in {index_file}")

    shard_names = sorted(set(weight_map.values()))
    shard_paths: list[Path] = []

    for name in shard_names:
        if not isinstance(name, str) or not name:
            raise RuntimeError(
                f"invalid safetensors index: shard name must be a non-empty string in {index_file}"
            )
        shard_rel = Path(name)
        if shard_rel.is_absolute() or ".." in shard_rel.parts:
            raise RuntimeError(
                f"invalid safetensors index: unsafe shard path {name!r} in {index_file}"
            )
        shard_path = base_dir / name
        if not shard_path.exists():
            raise RuntimeError(
                f"safetensors index references missing shard {name!r} in {base_dir}"
            )
        shard_paths.append(shard_path)

    return shard_paths


def load_state_dict_from_file(
    path: Path,
    global_state_dict: StateDictLike,
    merge_lock: Lock | None = None,
) -> int:
    suffix = path.suffix.lower()
    if suffix == ".safetensors":
        logger.info("Using safetensors instruments on %s", path)
        loaded = load_safetensors_file(str(path), device="cpu")
    else:
        logger.info("Using torch instruments on %s", path)
        loaded = torch.load(path, map_location="cpu")
        if isinstance(loaded, dict) and "state_dict" in loaded and isinstance(loaded["state_dict"], dict):
            logger.info("Detected wrapped state_dict payload in %s", path)
            loaded = loaded["state_dict"]
    loaded = validate_state_dict_mapping(loaded, path)
    loaded_count = 0
    for key, tensor in loaded.items():
        lock_context = merge_lock if merge_lock is not None else nullcontext()
        with lock_context:
            if key in global_state_dict:
                raise RuntimeError(f"duplicate tensor key {key!r} while loading file {path}")
            global_state_dict[key] = tensor
            loaded_count += 1
    return loaded_count


def validate_state_dict_mapping(loaded: object, path: Path) -> Dict[str, torch.Tensor]:
    if not isinstance(loaded, dict):
        raise RuntimeError(f"checkpoint at {path} is not a state_dict mapping")
    if not all(isinstance(k, str) and torch.is_tensor(v) for k, v in loaded.items()):
        raise RuntimeError(f"checkpoint at {path} is not a plain tensor state_dict")
    return dict(loaded)



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, TypeVar

from .refs import TensorRef, parse_slice
from .transform import (
    BinaryMappingSpec,
    BinaryMappingTransform,
    DestinationPolicy,
    ResolvedMapping,
    ResolvedTernaryMapping,
    TernaryMappingSpec,
    TernaryMappingTransform,
    UnarySpec,
    UnaryTransform,
)
from .types import StateDictProvider, TransformError
from .validation import ensure_mapping_payload, validate_payload_keys

UnarySpecT = TypeVar("UnarySpecT", bound=UnarySpec)
BinarySpecT = TypeVar("BinarySpecT", bound=BinaryMappingSpec)
TernarySpecT = TypeVar("TernarySpecT", bound=TernaryMappingSpec)


@dataclass(frozen=True)
class Docs:
    summary: str
    notes: tuple[str, ...] = ()
    examples: tuple[str, ...] = ()


@dataclass(frozen=True)
class UnaryRefs:
    target_slice: bool = False


@dataclass(frozen=True)
class BinaryRefs:
    from_slice: bool = False
    to_slice: bool = False


@dataclass(frozen=True)
class TernaryRefs:
    from_a_slice: bool = False
    from_b_slice: bool = False
    to_slice: bool = False


def _lines(text: Docs, *rules: str) -> str:
    lines = [text.summary, "", *rules]
    if text.notes:
        lines.extend(["", *text.notes])
    if text.examples:
        lines.extend(["", "Examples:", *[f"  {example}" for example in text.examples]])
    return "\n".join(lines)


def _slice_rule(label: str, allowed: bool) -> str:
    return (
        f"{label} may include slicing." if allowed else f"{label} must not be sliced."
    )


def _destination_rule(policy: DestinationPolicy) -> str:
    if policy is DestinationPolicy.MUST_EXIST:
        return "Destination tensors must already exist."
    if policy is DestinationPolicy.MUST_NOT_EXIST:
        return "Destination tensors must not already exist."
    return "Destination tensors may be created or overwritten."


def _validate_slice(
    ref: TensorRef,
    *,
    allowed: bool,
    op_name: str,
    label: str,
    error_type: type[TransformError],
) -> None:
    if ref.slice_spec is None:
        return
    if not allowed:
        raise error_type(f"{op_name} {label} must not be sliced")
    parse_slice(ref.slice_spec)


class DeclarativeUnaryTransform(UnaryTransform[UnarySpecT]):
    spec_type = UnarySpec
    allowed_keys = {"target"}
    required_keys = {"target"}
    docs: Docs
    refs = UnaryRefs()
    spec_builder: Callable[[TensorRef, dict], UnarySpecT] | None = None
    progress_desc: str | None = None
    apply_fn: Callable[[UnarySpecT, str, StateDictProvider], None]

    def __init_subclass__(cls, **kwargs: object) -> None:
        super().__init_subclass__(**kwargs)
        if getattr(cls, "docs", None) is None:
            return
        if cls.progress_desc is None:
            cls.progress_desc = f"Applying {cls.name} transforms"
        cls.slice_policy = "allow" if cls.refs.target_slice else "forbid"
        cls.help_text = _lines(
            cls.docs, _slice_rule("Target tensors", cls.refs.target_slice)
        )

    def build_spec(self, target_ref: TensorRef, payload: dict) -> UnarySpecT:
        if self.spec_builder is not None:
            return self.spec_builder(target_ref, payload)
        return self.spec_type(target_ref=target_ref)

    def apply_to_target(
        self, spec: UnarySpecT, name: str, provider: StateDictProvider
    ) -> None:
        self.apply_fn(spec, name, provider)


class DeclarativeBinaryTransform(BinaryMappingTransform[BinarySpecT]):
    spec_type = BinaryMappingSpec
    allowed_keys = {"from", "to"}
    required_keys = {"from", "to"}
    docs: Docs
    refs = BinaryRefs()
    spec_builder: Callable[[TensorRef, TensorRef, dict], BinarySpecT] | None = None
    progress_desc: str | None = None
    apply_fn: Callable[[BinarySpecT, ResolvedMapping, StateDictProvider], None]

    def __init_subclass__(cls, **kwargs: object) -> None:
        super().__init_subclass__(**kwargs)
        if getattr(cls, "docs", None) is None:
            return
        if cls.progress_desc is None:
            cls.progress_desc = f"Applying {cls.name} transforms"
        cls.help_text = _lines(
            cls.docs,
            _slice_rule("Source references", cls.refs.from_slice),
            _slice_rule("Destination references", cls.refs.to_slice),
            _destination_rule(cls.destination_policy),
        )

    def compile(self, payload: dict, default_model: str | None) -> BinarySpecT:
        if self.spec_builder is None:
            return super().compile(payload, default_model)
        payload = ensure_mapping_payload(payload, self.name)
        validate_payload_keys(
            payload,
            op_name=self.name,
            allowed_keys=self.allowed_keys,
            required_keys=self.required_keys,
        )
        from_ref, to_ref = self.parse_refs(payload, default_model)
        self.validate_refs(from_ref, to_ref)
        assert from_ref.model is not None
        assert to_ref.model is not None
        return self.spec_builder(from_ref, to_ref, payload)

    def validate_refs(self, from_ref: TensorRef, to_ref: TensorRef) -> None:
        _validate_slice(
            from_ref,
            allowed=self.refs.from_slice,
            op_name=self.name,
            label="source",
            error_type=self.error_type,
        )
        _validate_slice(
            to_ref,
            allowed=self.refs.to_slice,
            op_name=self.name,
            label="destination",
            error_type=self.error_type,
        )

    def apply_item(
        self, spec: BinarySpecT, item: ResolvedMapping, provider: StateDictProvider
    ) -> None:
        self.apply_fn(spec, item, provider)

    def apply_mapping(self, item: ResolvedMapping, provider: StateDictProvider) -> None:
        raise NotImplementedError


class DeclarativeTernaryTransform(TernaryMappingTransform[TernarySpecT]):
    spec_type = TernaryMappingSpec
    allowed_keys = {"from_a", "from_b", "to"}
    required_keys = {"from_a", "from_b", "to"}
    docs: Docs
    refs = TernaryRefs()
    progress_desc: str | None = None
    apply_fn: Callable[[TernarySpecT, ResolvedTernaryMapping, StateDictProvider], None]

    def __init_subclass__(cls, **kwargs: object) -> None:
        super().__init_subclass__(**kwargs)
        if getattr(cls, "docs", None) is None:
            return
        if cls.progress_desc is None:
            cls.progress_desc = f"Applying {cls.name} transforms"
        cls.help_text = _lines(
            cls.docs,
            "References may be regex or structured mappings.",
            _slice_rule("'from_a' references", cls.refs.from_a_slice),
            _slice_rule("'from_b' references", cls.refs.from_b_slice),
            _slice_rule("'to' references", cls.refs.to_slice),
            _destination_rule(cls.destination_policy),
        )

    def validate_refs(
        self, from_a_ref: TensorRef, from_b_ref: TensorRef, to_ref: TensorRef
    ) -> None:
        _validate_slice(
            from_a_ref,
            allowed=self.refs.from_a_slice,
            op_name=self.name,
            label="from_a",
            error_type=self.error_type,
        )
        _validate_slice(
            from_b_ref,
            allowed=self.refs.from_b_slice,
            op_name=self.name,
            label="from_b",
            error_type=self.error_type,
        )
        _validate_slice(
            to_ref,
            allowed=self.refs.to_slice,
            op_name=self.name,
            label="destination",
            error_type=self.error_type,
        )

    def apply_item(
        self,
        spec: TernarySpecT,
        item: ResolvedTernaryMapping,
        provider: StateDictProvider,
    ) -> None:
        self.apply_fn(spec, item, provider)

    def apply_mapping(
        self, item: ResolvedTernaryMapping, provider: StateDictProvider
    ) -> None:
        raise NotImplementedError


__all__ = [
    "Docs",
    "UnaryRefs",
    "BinaryRefs",
    "TernaryRefs",
    "DeclarativeUnaryTransform",
    "DeclarativeBinaryTransform",
    "DeclarativeTernaryTransform",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from ..core import Expression, TransformError, collect_ref_models, compile_tensor_ref_expr, format_ref, register_assert_expr, resolve_matches
from ..core import TensorRef
from ..core import StateDictProvider


@dataclass(frozen=True)
class ExistsExpr:
    ref: TensorRef

    def evaluate(self, provider: StateDictProvider) -> None:
        matches = resolve_matches(self.ref, provider, op_name="exists")
        if not matches:
            raise TransformError(f"exists failed: {format_ref(self.ref)} matched zero tensors")

    def collect_models(self) -> set[str]:
        return collect_ref_models(self.ref)


@register_assert_expr(
    "exists",
    payload_kind="tensor-ref",
    description="Succeeds if the reference matches at least one tensor.",
)
def compile_exists_expr(payload: Any, default_model: str | None) -> Expression:
    return ExistsExpr(ref=compile_tensor_ref_expr(payload, default_model, "exists"))



from __future__ import annotations

from dataclasses import dataclass

from ..core import BinaryMappingSpec, DestinationPolicy, UnarySpec
import torch
from ..core import ResolvedMapping, StateDictProvider, TensorRef, TransformError, must_model, select_tensor
from ..core import register_transform
from ..core import require_nonempty_string
from ..core import (
    parse_torch_dtype,
)
from ..core import BinaryRefs, DeclarativeBinaryTransform, Docs, DeclarativeUnaryTransform, UnaryRefs


@dataclass(frozen=True)
class CastSpec(BinaryMappingSpec):
    dtype: torch.dtype


def _build_cast_spec(from_ref: TensorRef, to_ref: TensorRef, payload: dict) -> CastSpec:
    raw_dtype = require_nonempty_string(payload, op_name="cast", key="dtype")
    dtype = parse_torch_dtype(
        raw_dtype,
        error_type=TransformError,
        op_name="cast",
        field_name="dtype",
    )
    return CastSpec(from_ref=from_ref, to_ref=to_ref, dtype=dtype)


def _cast_apply(
    spec: CastSpec, item: ResolvedMapping, provider: StateDictProvider
) -> None:
    src_sd = provider.get_state_dict(item.src_model)
    dst_sd = provider.get_state_dict(item.dst_model)
    src_view = select_tensor(src_sd[item.src_name], item.src_slice)
    dst_sd[item.dst_name] = src_view.to(dtype=spec.dtype)


class CastTransform(DeclarativeBinaryTransform[CastSpec]):
    name = "cast"
    error_type = TransformError
    spec_type = CastSpec
    destination_policy = DestinationPolicy.MUST_NOT_EXIST
    allowed_keys = {"from", "to", "dtype"}
    required_keys = {"from", "to", "dtype"}
    docs = Docs(
        "Casts source tensors to a different dtype and writes new destination tensors.",
        examples=(
            "cast: { from: ln_f.weight, to: ln_f_fp16.weight, dtype: float16 }",
            "cast: { from: '.*weight', to: 'fp16.\\\\g<0>', dtype: bfloat16 }",
        ),
    )
    refs = BinaryRefs(from_slice=True)
    spec_builder = staticmethod(_build_cast_spec)
    apply_fn = staticmethod(_cast_apply)


register_transform(CastTransform())


@dataclass(frozen=True)
class CastInPlaceSpec(UnarySpec):
    dtype: torch.dtype


def _build_cast_in_place_spec(target_ref: TensorRef, payload: dict) -> CastInPlaceSpec:
    raw_dtype = require_nonempty_string(payload, op_name="cast_", key="to")
    dtype = parse_torch_dtype(
        raw_dtype,
        error_type=TransformError,
        op_name="cast_",
        field_name="to",
    )
    return CastInPlaceSpec(target_ref=target_ref, dtype=dtype)


def _cast_in_place_apply(
    spec: CastInPlaceSpec, name: str, provider: StateDictProvider
) -> None:
    model = must_model(spec.target_ref)
    sd = provider.get_state_dict(model)
    sd[name] = sd[name].to(dtype=spec.dtype)


class CastInPlaceTransform(DeclarativeUnaryTransform[CastInPlaceSpec]):
    name = "cast_"
    error_type = TransformError
    spec_type = CastInPlaceSpec
    allowed_keys = {"target", "to"}
    required_keys = {"target", "to"}
    docs = Docs(
        "Casts one or more tensors to a different dtype in-place.",
        notes=("The entire tensor is cast.",),
        examples=(
            "cast_: { target: ln_f.weight, to: float16 }",
            "cast_: { target: '.*weight', to: bfloat16 }",
        ),
    )
    refs = UnaryRefs()
    spec_builder = staticmethod(_build_cast_in_place_spec)
    apply_fn = staticmethod(_cast_in_place_apply)


register_transform(CastInPlaceTransform())



from __future__ import annotations

from pathlib import Path
from typing import Any, Literal

import numpy as np
import torch
from safetensors.torch import load_file as load_safetensors_file
from safetensors.torch import save_file as save_safetensors_file


def infer_tensor_file_format(path: Path) -> Literal["numpy", "safetensors", "torch"]:
    suffix = path.suffix.lower()
    if suffix in {".npy", ".npz"}:
        return "numpy"
    if suffix == ".safetensors":
        return "safetensors"
    return "torch"


def load_tensor_from_path(
    path: Path,
    *,
    format: Literal["auto", "numpy", "safetensors", "torch"] = "auto",
) -> torch.Tensor:
    format_name = infer_tensor_file_format(path) if format == "auto" else format

    if format_name == "numpy":
        loaded = np.load(path, allow_pickle=False)
        if isinstance(loaded, np.ndarray):
            return torch.from_numpy(loaded)
        if isinstance(loaded, np.lib.npyio.NpzFile):
            keys = list(loaded.keys())
            if len(keys) != 1:
                raise RuntimeError(
                    "loading tensor from .npz requires exactly one array, or use state_dict load"
                )
            return torch.from_numpy(loaded[keys[0]])
        raise RuntimeError(f"unsupported numpy payload at {path}")

    if format_name == "safetensors":
        loaded = load_safetensors_file(str(path), device="cpu")
        if len(loaded) != 1:
            raise RuntimeError(
                "loading tensor from safetensors requires exactly one tensor, or use state_dict load"
            )
        return next(iter(loaded.values()))

    loaded_obj: Any = torch.load(path, map_location="cpu")
    if torch.is_tensor(loaded_obj):
        return loaded_obj
    if isinstance(loaded_obj, np.ndarray):
        return torch.from_numpy(loaded_obj)
    if isinstance(loaded_obj, dict):
        candidate = loaded_obj.get("state_dict", loaded_obj)
        if isinstance(candidate, dict):
            tensors = [
                (k, v) for k, v in candidate.items() if isinstance(k, str) and torch.is_tensor(v)
            ]
            if len(tensors) == 1:
                return tensors[0][1]
            raise RuntimeError(
                "loading tensor from torch checkpoint requires exactly one tensor, or use state_dict load"
            )
    raise RuntimeError(f"unsupported tensor payload at {path}")


def save_tensor_to_path(
    tensor_name: str,
    tensor: torch.Tensor,
    path: Path,
    *,
    format: Literal["safetensors", "torch", "numpy"],
) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    if format == "torch":
        torch.save(tensor, path)
        return
    if format == "numpy":
        np.save(path, tensor.detach().cpu().numpy())
        return
    save_safetensors_file({tensor_name: tensor}, str(path))



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import torch

from ..core import match_expr_names
from ..core import TensorRef, must_model, parse_model_expr, parse_slice, select_tensor
from ..core import StateDictProvider, TransformError
from ..core import BaseTransform, TransformResult, register_transform
from ..core import ensure_mapping_payload, validate_payload_keys


class SplitTransformError(TransformError):
    pass


@dataclass(frozen=True)
class SplitSpec:
    from_ref: TensorRef
    to_refs: list[TensorRef]
    sizes: list[int]
    dim: int

    def collect_models(self) -> set[str]:
        models = {must_model(self.from_ref)}
        models.update(must_model(ref) for ref in self.to_refs)
        return models


class SplitTransform(BaseTransform):
    name = "split"
    error_type = SplitTransformError
    spec_type = SplitSpec
    allowed_keys = {"from", "to", "sizes", "dim"}
    required_keys = {"from", "to", "sizes"}
    help_text = (
        "Splits one source tensor into multiple destination tensors.\n"
        "\n"
        "'sizes' must sum to the source size along 'dim'. Destination refs must be\n"
        "single unsliced tensor names and must not already exist.\n"
        "\n"
        "Example:\n"
        "  split: { from: x, to: [x0, x1], sizes: [32, 32], dim: 0 }"
    )

    def completion_reference_keys(self) -> list[str]:
        return ["from", "to"]

    def compile(self, payload: Any, default_model: str | None) -> SplitSpec:
        payload = ensure_mapping_payload(payload, self.name)
        validate_payload_keys(
            payload,
            op_name=self.name,
            allowed_keys=self.allowed_keys,
            required_keys=self.required_keys,
        )
        from_ref = parse_model_expr(payload.get("from"), default_model=default_model)
        if from_ref.slice_spec is not None:
            parse_slice(from_ref.slice_spec)
        if from_ref.model is None:
            raise SplitTransformError("split.from missing model alias")

        raw_to = payload.get("to")
        if not isinstance(raw_to, list) or len(raw_to) < 2:
            raise SplitTransformError("split.to must be a list of at least two references")
        to_refs: list[TensorRef] = []
        for idx, item in enumerate(raw_to):
            ref = parse_model_expr(item, default_model=default_model)
            if ref.model is None:
                raise SplitTransformError(f"split.to[{idx}] missing model alias")
            if ref.slice_spec is not None:
                raise SplitTransformError("split destination references must not be sliced")
            if not isinstance(ref.expr, str):
                raise SplitTransformError("split destination references must resolve to single names")
            to_refs.append(ref)

        sizes = _parse_sizes(payload.get("sizes"))
        if len(sizes) != len(to_refs):
            raise SplitTransformError("split.sizes length must match split.to length")

        raw_dim = payload.get("dim", 0)
        if not isinstance(raw_dim, int):
            raise SplitTransformError("split.dim must be an integer")

        return SplitSpec(from_ref=from_ref, to_refs=to_refs, sizes=sizes, dim=raw_dim)

    def apply(self, spec: object, provider: StateDictProvider) -> TransformResult:
        typed = self.require_spec(spec)
        src_model = must_model(typed.from_ref)
        src_sd = provider.get_state_dict(src_model)
        matches = match_expr_names(
            expr=typed.from_ref.expr,
            names=src_sd.keys(),
            op_name=self.name,
            role="source",
        )
        if not matches:
            raise SplitTransformError("split source matched zero tensors")
        if len(matches) != 1:
            raise SplitTransformError(f"split source must match exactly one tensor, got {len(matches)}")
        src_name = matches[0]
        src_slice = parse_slice(typed.from_ref.slice_spec) if typed.from_ref.slice_spec is not None else None
        src_view = select_tensor(src_sd[src_name], src_slice)

        rank = src_view.dim()
        dim = typed.dim if typed.dim >= 0 else typed.dim + rank
        if dim < 0 or dim >= rank:
            raise SplitTransformError(f"split.dim {typed.dim} out of range for rank {rank} tensor")
        if sum(typed.sizes) != int(src_view.shape[dim]):
            raise SplitTransformError(
                f"split.sizes must sum to source size along dim {dim}: "
                f"{sum(typed.sizes)} != {int(src_view.shape[dim])}"
            )

        parts = torch.split(src_view, typed.sizes, dim=dim)
        for ref, part in zip(typed.to_refs, parts, strict=True):
            dst_model = must_model(ref)
            assert isinstance(ref.expr, str)
            dst_name = ref.expr
            dst_sd = provider.get_state_dict(dst_model)
            if dst_name in dst_sd:
                raise SplitTransformError(f"split destination already exists: {dst_model}::{dst_name}")
            dst_sd[dst_name] = part.clone()

        return TransformResult(name=self.name, count=len(parts))

    def infer_output_model(self, spec: object) -> str:
        typed = self.require_spec(spec)
        return must_model(typed.to_refs[0])

    def require_spec(self, spec: object) -> SplitSpec:
        if not isinstance(spec, SplitSpec):
            raise SplitTransformError(f"split received wrong spec type: {type(spec).__name__}")
        return spec


def _parse_sizes(raw: object) -> list[int]:
    if not isinstance(raw, list) or not raw:
        raise SplitTransformError("split.sizes must be a non-empty list of positive integers")
    if not all(isinstance(x, int) and x > 0 for x in raw):
        raise SplitTransformError("split.sizes must be a non-empty list of positive integers")
    return list(raw)








register_transform(SplitTransform())



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from ..expressions import get_assert_expr_help, get_assert_expr_names
from ..core import TransformError
from ..core import TypedTransform, TransformControl, TransformResult, get_transform, list_transforms, register_transform
from ..core import StateDictProvider
from ..engine.frontend import emit_line


class HelpTransformError(TransformError):
    pass


@dataclass(frozen=True)
class HelpSpec:
    command: str | None = None
    subcommand: str | None = None

    def collect_models(self) -> set[str]:
        return set()


class HelpTransform(TypedTransform[HelpSpec]):
    name = "help"
    error_type = HelpTransformError
    spec_type = HelpSpec
    help_text = (
        "Shows help for commands and assert expressions.\n"
        "\n"
        "Run without arguments to list all commands. You can request help for a specific "
        "command or, for 'assert', for an individual expression operator.\n"
        "\n"
        "Examples:\n"
        "  help\n"
        "  help: copy\n"
        "  help: assert\n"
        "  help: { assert: equal }"
    )

    def compile(self, payload: Any, default_model: str | None) -> HelpSpec:
        del default_model

        if payload is None:
            return HelpSpec(command=None, subcommand=None)

        if isinstance(payload, str):
            cmd = payload.strip()
            if not cmd:
                raise HelpTransformError("help payload must be a non-empty string")
            return HelpSpec(command=cmd, subcommand=None)

        if isinstance(payload, dict):
            if not payload:
                return HelpSpec(command=None, subcommand=None)

            if len(payload) != 1:
                raise HelpTransformError("help mapping payload must have exactly one key")

            command, subpayload = next(iter(payload.items()))
            if not isinstance(command, str) or not command.strip():
                raise HelpTransformError("help command must be a non-empty string")

            command = command.strip()

            if subpayload is None:
                return HelpSpec(command=command, subcommand=None)

            if not isinstance(subpayload, str) or not subpayload.strip():
                raise HelpTransformError(
                    "help subcommand must be a non-empty string when provided"
                )

            return HelpSpec(command=command, subcommand=subpayload.strip())

        raise HelpTransformError("help payload must be empty, a string, or a single-key mapping")

    def apply(self, spec: object, provider: StateDictProvider) -> TransformResult:
        del provider

        typed = self.require_spec(spec)

        if typed.command is None:
            self._print_all_commands()
        elif typed.command == "assert":
            if typed.subcommand is None:
                self._print_assert_help()
            else:
                self._print_assert_expr_help(typed.subcommand)
        else:
            if typed.subcommand is not None:
                raise HelpTransformError(
                    f"command {typed.command!r} does not support subcommand help"
                )
            self._print_command_help(typed.command)

        return TransformResult(
            name=self.name,
            count=0,
            control=TransformControl.CONTINUE,
        )

    def infer_output_model(self, spec: object) -> str:
        del spec
        raise HelpTransformError("help does not infer an output model")

    def contributes_output_model(self, spec: object) -> bool:
        del spec
        return False

    def completion_payload_start_candidates(self, prefix_text: str) -> list[str] | None:
        candidates = [name for name in list_transforms() if name.startswith(prefix_text)]
        if not prefix_text:
            return candidates + ["{ "]
        return candidates

    def completion_key_candidates(self, before_cursor: str, prefix_text: str) -> list[str] | None:
        del before_cursor
        candidate = "assert: "
        if candidate.startswith(prefix_text):
            return [candidate]
        return []

    def completion_value_candidates(
        self,
        value_key: str | None,
        prefix_text: str,
        model_aliases: list[str],
    ) -> list[str] | None:
        del model_aliases
        if value_key == "help":
            return [name for name in list_transforms() if name.startswith(prefix_text)]
        if value_key == "assert":
            return [name for name in get_assert_expr_names() if name.startswith(prefix_text)]
        return None

    def completion_committed_next_candidates(self, value_key: str | None) -> list[str] | None:
        if value_key in {"help", "assert"}:
            return ["}"]
        return None

    def _print_all_commands(self) -> None:
        emit_line("Available commands:")
        for name in list_transforms():
            emit_line(f"  {name}")
        emit_line("")
        emit_line("For help on a specific command, run: help: <command>")
        emit_line("For help on a specific assert expression, run: help: { assert: <expr> }")

    def _print_command_help(self, command_name: str) -> None:
        try:
            transform = get_transform(command_name)
        except TransformError as exc:
            raise HelpTransformError(f"unknown command: {command_name}") from exc

        allowed_keys = getattr(transform, "allowed_keys", None)
        required_keys = getattr(transform, "required_keys", None)
        help_text = getattr(transform, "help_text", None)

        emit_line(f"Command: {command_name}")

        if help_text:
            emit_line(help_text)

        if allowed_keys is None and required_keys is None:
            emit_line("Key metadata: unavailable")
            return

        allowed = sorted(allowed_keys or set())
        required = sorted(required_keys or set())

        if required:
            emit_line("Required keys:")
            for key in required:
                emit_line(f"  - {key}")
        else:
            emit_line("Required keys: none")

        optional = [key for key in allowed if key not in required]
        if optional:
            emit_line("Optional keys:")
            for key in optional:
                emit_line(f"  - {key}")
        else:
            emit_line("Optional keys: none")

        if allowed:
            emit_line("All allowed keys:")
            for key in allowed:
                emit_line(f"  - {key}")
        else:
            emit_line("All allowed keys: none")

    def _print_assert_help(self) -> None:
        try:
            transform = get_transform("assert")
        except TransformError:
            transform = None

        emit_line("Command: assert")
        if transform is not None:
            help_text = getattr(transform, "help_text", None)
            if help_text:
                emit_line(help_text)

        emit_line("Supported assert expression operators:")
        for name in get_assert_expr_names():
            meta = get_assert_expr_help(name)
            if meta.description:
                emit_line(f"  {name}: {meta.description}")
            else:
                emit_line(f"  {name}")

        emit_line("")
        emit_line("For help on a specific assert expression, run: help: { assert: <expr> }")

    def _print_assert_expr_help(self, expr_name: str) -> None:
        meta = get_assert_expr_help(expr_name)

        emit_line(f"Assert expression: {meta.name}")
        emit_line(f"Payload: {meta.payload_kind}")

        if meta.description:
            emit_line(meta.description)

        required = sorted(meta.required_keys or [])
        allowed = sorted(meta.allowed_keys or [])
        optional = [key for key in allowed if key not in required]

        if meta.required_keys is not None:
            if required:
                emit_line("Required keys:")
                for key in required:
                    emit_line(f"  - {key}")
            else:
                emit_line("Required keys: none")

        if meta.allowed_keys is not None:
            if optional:
                emit_line("Optional keys:")
                for key in optional:
                    emit_line(f"  - {key}")
            else:
                emit_line("Optional keys: none")

            if allowed:
                emit_line("All allowed keys:")
                for key in allowed:
                    emit_line(f"  - {key}")
            else:
                emit_line("All allowed keys: none")










register_transform(HelpTransform())



from __future__ import annotations

import json
from typing import Any

import torch


def summarize_tensor(
    tensor: torch.Tensor,
    *,
    verbosity: str,
    access_counts: dict[str, int] | None = None,
) -> dict[str, Any]:
    t = tensor.detach()

    if verbosity == "shape":
        return {"shape": list(t.shape)}

    if verbosity == "full":
        values_tensor = t.cpu()
        summary = {
            "shape": list(t.shape),
            "dtype": str(t.dtype),
            "device": str(t.device),
            "values": values_tensor.tolist(),
        }
        if access_counts is not None:
            summary["reads"] = access_counts["reads"]
            summary["writes"] = access_counts["writes"]
        return summary

    if t.numel() == 0:
        summary = {
            "shape": list(t.shape),
            "min": None,
            "max": None,
            "mean": None,
        }
        if access_counts is not None:
            summary["reads"] = access_counts["reads"]
            summary["writes"] = access_counts["writes"]
        return summary

    stat_tensor = t
    if not stat_tensor.is_floating_point():
        stat_tensor = stat_tensor.to(torch.float32)

    summary = {
        "shape": list(t.shape),
        "min": float(stat_tensor.min().item()),
        "max": float(stat_tensor.max().item()),
        "mean": float(stat_tensor.mean().item()),
    }
    if access_counts is not None:
        summary["reads"] = access_counts["reads"]
        summary["writes"] = access_counts["writes"]
    return summary


def _is_tensor_summary(node: Any) -> bool:
    return (
        isinstance(node, dict)
        and set(node.keys())
        in (
            {"shape"},
            {"shape", "min", "max", "mean"},
            {"shape", "min", "max", "mean", "reads", "writes"},
            {"shape", "dtype", "device", "values"},
            {"shape", "dtype", "device", "values", "reads", "writes"},
        )
    )


def _shape_only(node: Any) -> Any:
    if _is_tensor_summary(node):
        return {"shape": node["shape"]}

    if isinstance(node, dict):
        return {key: _shape_only(value) for key, value in node.items()}

    if isinstance(node, list):
        return [None if value is None else _shape_only(value) for value in node]

    return node


def _format_summary(summary: dict[str, Any], *, compact: bool) -> str:
    del compact  # layout is handled by the tree renderer, not by summary content

    shape = summary["shape"]

    if set(summary.keys()) == {"shape"}:
        return f"shape={shape}"

    if {"dtype", "device", "values"}.issubset(summary.keys()):
        rendered = (
            f"shape={shape} "
            f"dtype={summary['dtype']} "
            f"device={summary['device']} "
            f"values={summary['values']!r}"
        )
        if "reads" in summary and "writes" in summary:
            rendered += f" reads={summary['reads']} writes={summary['writes']}"
        return rendered

    min_value = summary["min"]
    max_value = summary["max"]
    mean_value = summary["mean"]

    if min_value is None:
        rendered = f"shape={shape} min=None max=None mean=None"
        if "reads" in summary and "writes" in summary:
            rendered += f" reads={summary['reads']} writes={summary['writes']}"
        return rendered

    rendered = (
        f"shape={shape} "
        f"min={min_value:.6g} "
        f"max={max_value:.6g} "
        f"mean={mean_value:.6g}"
    )
    if "reads" in summary and "writes" in summary:
        rendered += f" reads={summary['reads']} writes={summary['writes']}"
    return rendered


def render_tree(tree: dict[str, Any], *, compact: bool) -> str:
    lines: list[str] = []
    items = list(tree.items())

    for index, (key, value) in enumerate(items):
        lines.extend(
            _render_node(
                str(key),
                value,
                prefix="",
                is_last=index == len(items) - 1,
                compact=compact,
            )
        )

    return "\n".join(lines)


def _render_node(
    name: str,
    node: Any,
    *,
    prefix: str,
    is_last: bool,
    compact: bool,
) -> list[str]:
    branch = "└── " if is_last else "├── "
    child_prefix = prefix + ("    " if is_last else "│   ")

    if _is_tensor_summary(node):
        return [f"{prefix}{branch}{name}  {_format_summary(node, compact=compact)}"]

    lines = [f"{prefix}{branch}{name}"]

    if isinstance(node, dict):
        items = list(node.items())
        for index, (child_name, child_node) in enumerate(items):
            lines.extend(
                _render_node(
                    str(child_name),
                    child_node,
                    prefix=child_prefix,
                    is_last=index == len(items) - 1,
                    compact=compact,
                )
            )
        return lines

    if isinstance(node, list):
        groups = _list_group_entries(node, compact=compact)
        for index, (label, child_node) in enumerate(groups):
            lines.extend(
                _render_node(
                    label,
                    child_node,
                    prefix=child_prefix,
                    is_last=index == len(groups) - 1,
                    compact=compact,
                )
            )
        return lines

    return [f"{prefix}{branch}{name}  {node!r}"]


def _list_group_entries(node: list[Any], *, compact: bool) -> list[tuple[str, Any]]:
    present = [(index, value) for index, value in enumerate(node) if value is not None]
    if not present:
        return []

    if not compact:
        return [(_format_index_range(index, index), value) for index, value in present]

    groups: list[tuple[int, int, Any]] = []

    start, first_value = present[0]
    prev = start
    current_value = first_value
    current_key = _canonical_key(first_value)

    for index, value in present[1:]:
        value_key = _canonical_key(value)

        if index == prev + 1 and value_key == current_key:
            prev = index
            continue

        groups.append((start, prev, current_value))
        start = index
        prev = index
        current_value = value
        current_key = value_key

    groups.append((start, prev, current_value))

    return [(_format_index_range(start, end), value) for start, end, value in groups]


def _format_index_range(start: int, end: int) -> str:
    if start == end:
        return f"[{start}]"
    return f"[{start}-{end}]"


def _canonical_key(node: Any) -> str:
    return json.dumps(node, sort_keys=True, separators=(",", ":"))


__all__ = [
    "summarize_tensor",
    "render_tree",
]



from __future__ import annotations

import torch

from ..core import TransformError


class PhloraSvdCache:
    def __init__(self) -> None:
        self._cache: dict[str, tuple[torch.Tensor, torch.Tensor, torch.Tensor]] = {}

    def get(
        self,
        source: torch.Tensor,
        *,
        cache_key: str,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        full_key = f"{cache_key}|{tuple(source.shape)}|{source.dtype}|{source.device}"
        if full_key not in self._cache:
            self._cache[full_key] = torch.linalg.svd(source, full_matrices=False)
        return self._cache[full_key]


def require_positive_rank(
    value: float,
    *,
    error_type: type[TransformError],
    op_name: str,
    key: str,
) -> int:
    integer = int(value)
    if float(integer) != float(value) or integer <= 0:
        raise error_type(f"{op_name}.{key} must be a positive integer")
    return integer


def _resolve_effective_rank(
    source: torch.Tensor,
    requested_rank: int,
    *,
    error_type: type[TransformError],
    op_name: str,
    tensor_name: str,
) -> int:
    _require_matrix(
        source,
        error_type=error_type,
        op_name=op_name,
        tensor_name=tensor_name,
    )
    rank = min(requested_rank, min(source.shape))
    if rank <= 0:
        raise error_type(
            f"{op_name} rank became zero for {tensor_name} with shape {tuple(source.shape)}"
        )
    return rank


def _require_matrix(
    source: torch.Tensor,
    *,
    error_type: type[TransformError],
    op_name: str,
    tensor_name: str,
) -> None:
    if source.ndim != 2:
        raise error_type(
            f"{op_name} target must be 2D (got shape {tuple(source.shape)}): {tensor_name}"
        )


def compute_phlora_factors(
    source: torch.Tensor,
    requested_rank: int,
    *,
    cache: PhloraSvdCache,
    cache_key: str,
    error_type: type[TransformError],
    op_name: str,
    tensor_name: str,
) -> tuple[torch.Tensor, torch.Tensor]:
    rank = _resolve_effective_rank(
        source,
        requested_rank,
        error_type=error_type,
        op_name=op_name,
        tensor_name=tensor_name,
    )
    u, s, vh = cache.get(source, cache_key=cache_key)
    sqrt_s = s[:rank].sqrt()
    lora_a = sqrt_s[:, None] * vh[:rank, :]
    lora_b = u[:, :rank] * sqrt_s
    return lora_a, lora_b


def reconstruct_phlora_rank(
    source: torch.Tensor,
    requested_rank: int,
    *,
    cache: PhloraSvdCache,
    cache_key: str,
    error_type: type[TransformError],
    op_name: str,
    tensor_name: str,
) -> torch.Tensor:
    rank = _resolve_effective_rank(
        source,
        requested_rank,
        error_type=error_type,
        op_name=op_name,
        tensor_name=tensor_name,
    )
    u, s, vh = cache.get(source, cache_key=cache_key)
    return (u[:, :rank] * s[:rank]) @ vh[:rank, :]


__all__ = [
    "PhloraSvdCache",
    "_require_matrix",
    "_resolve_effective_rank",
    "compute_phlora_factors",
    "reconstruct_phlora_rank",
    "require_positive_rank",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import torch

from ..core import collect_ref_models, compile_shape, compile_tensor_ref_expr, format_ref, register_assert_expr, resolve_tensors, require_mapping_assert_payload
from ..core import TransformError
from ..core import TensorRef
from ..core import StateDictProvider


@dataclass(frozen=True)
class ShapeExpr:
    ref: TensorRef
    is_value: tuple[int, ...]

    def evaluate(self, provider: StateDictProvider) -> None:
        for ref, tensor in resolve_tensors(self.ref, provider, op_name="shape.of"):
            if tuple(tensor.shape) != self.is_value:
                raise TransformError(
                    f"shape failed: {format_ref(ref)} has shape {tuple(tensor.shape)}, expected {self.is_value}"
                )

    def collect_models(self) -> set[str]:
        return collect_ref_models(self.ref)


@register_assert_expr(
    "shape",
    payload_kind="mapping",
    allowed_keys={"of", "is"},
    required_keys={"of", "is"},
    description="Succeeds if the tensor has the given shape.",
)
def compile_shape_expr(payload: Any, default_model: str | None) -> ShapeExpr:
    payload = require_mapping_assert_payload(
        payload,
        op_name="shape",
        allowed_keys={"of", "is"},
        required_keys={"of", "is"},
    )
    ref = compile_tensor_ref_expr(payload["of"], default_model, "shape.of")
    return ShapeExpr(ref=ref, is_value=compile_shape(payload["is"]))
